
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.protocol.http`` module contains the HttpRpc protocol
implementation.

This module is EXPERIMENTAL. You may not recognize the code here next time you
look at it.
"""

import logging
logger = logging.getLogger(__name__)

import re
import pytz
import tempfile

from spyne import BODY_STYLE_WRAPPED, MethodDescriptor, PushBase
from spyne.util import six, coroutine, Break
from spyne.util.six import string_types, BytesIO
from spyne.error import ResourceNotFoundError
from spyne.model.binary import BINARY_ENCODING_URLSAFE_BASE64, File
from spyne.model.primitive import DateTime
from spyne.protocol.dictdoc import SimpleDictDocument


TEMPORARY_DIR = None
STREAM_READ_BLOCK_SIZE = 0x4000
SWAP_DATA_TO_FILE_THRESHOLD = 512 * 1024



_OctalPatt = re.compile(r"\\[0-3][0-7][0-7]")
_QuotePatt = re.compile(r"[\\].")
_nulljoin = ''.join

def _unquote_cookie(str):
    """Handle double quotes and escaping in cookie values.
    This method is copied verbatim from the Python 3.5 standard
    library (http.cookies._unquote) so we don't have to depend on
    non-public interfaces.
    """
    # If there aren't any doublequotes,
    # then there can't be any special characters.  See RFC 2109.
    if str is None or len(str) < 2:
        return str
    if str[0] != '"' or str[-1] != '"':
        return str

    # We have to assume that we must decode this string.
    # Down to work.

    # Remove the "s
    str = str[1:-1]

    # Check for special sequences.  Examples:
    #    \012 --> \n
    #    \"   --> "
    #
    i = 0
    n = len(str)
    res = []
    while 0 <= i < n:
        o_match = _OctalPatt.search(str, i)
        q_match = _QuotePatt.search(str, i)
        if not o_match and not q_match:              # Neither matched
            res.append(str[i:])
            break
        # else:
        j = k = -1
        if o_match:
            j = o_match.start(0)
        if q_match:
            k = q_match.start(0)
        if q_match and (not o_match or k < j):     # QuotePatt matched
            res.append(str[i:k])
            res.append(str[k+1])
            i = k + 2
        else:                                      # OctalPatt matched
            res.append(str[i:j])
            res.append(chr(int(str[j+1:j+4], 8)))
            i = j + 4
    return _nulljoin(res)


def _parse_cookie(cookie):
    """Parse a ``Cookie`` HTTP header into a dict of name/value pairs.
    This function attempts to mimic browser cookie parsing behavior;
    it specifically does not follow any of the cookie-related RFCs
    (because browsers don't either).
    The algorithm used is identical to that used by Django version 1.9.10.
    """
    cookiedict = {}
    for chunk in cookie.split(str(';')):
        if str('=') in chunk:
            key, val = chunk.split(str('='), 1)
        else:
            # Assume an empty name per
            # https://bugzilla.mozilla.org/show_bug.cgi?id=169091
            key, val = str(''), chunk
        key, val = key.strip(), val.strip()
        if key or val:
            # unquote using Python's algorithm.
            cookiedict[key] = _unquote_cookie(val)
    return cookiedict


def get_stream_factory(dir=None, delete=True):
    def stream_factory(total_content_length, filename, content_type,
                                                           content_length=None):
        if total_content_length >= SWAP_DATA_TO_FILE_THRESHOLD or \
                                                                delete == False:
            if delete == False:
                # You need python >= 2.6 for this.
                retval = tempfile.NamedTemporaryFile('wb+', dir=dir,
                                                                  delete=delete)
            else:
                retval = tempfile.NamedTemporaryFile('wb+', dir=dir)
        else:
            retval = BytesIO()

        return retval

    return stream_factory

_weekday = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
_month = ['w00t', "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
             "Oct", "Nov", "Dec"]

def _header_to_bytes(prot, val, cls):
    if issubclass(cls, DateTime):
        if val.tzinfo is not None:
            val = val.astimezone(pytz.utc)
        else:
            val = val.replace(tzinfo=pytz.utc)

        return "%s, %02d %s %04d %02d:%02d:%02d GMT" % (
                            _weekday[val.weekday()], val.day, _month[val.month],
                            val.year, val.hour, val.minute, val.second)
    else:
        # because wsgi_ref wants header values in unicode.
        return prot.to_unicode(cls, val)


class HttpRpc(SimpleDictDocument):
    """The so-called HttpRpc protocol implementation. It only works with Http
    (wsgi and twisted) transports.

    :param app: An :class:'spyne.application.Application` instance.
    :param validator: Validation method to use. One of (None, 'soft')
    :param mime_type: Default mime type to set. Default is
        'application/octet-stream'
    :param tmp_dir: Temporary directory to store partial file uploads. Default
        is to use the OS default.
    :param tmp_delete_on_close: The ``delete`` argument to the
        :class:`tempfile.NamedTemporaryFile`.
        See: http://docs.python.org/2/library/tempfile.html#tempfile.NamedTemporaryFile.
    :param ignore_uncap: As HttpRpc can't serialize complex models, it throws a
        server exception when the return type of the user function is Complex.
        Passing ``True`` to this argument prevents that by ignoring the return
        value.
    """

    mime_type = 'text/plain'
    default_binary_encoding = BINARY_ENCODING_URLSAFE_BASE64
    default_string_encoding = 'UTF-8'

    type = set(SimpleDictDocument.type)
    type.add('http')

    def __init__(self, app=None, validator=None, mime_type=None,
                    tmp_dir=None, tmp_delete_on_close=True, ignore_uncap=False,
                        parse_cookie=True, hier_delim=".", strict_arrays=False):
        super(HttpRpc, self).__init__(app, validator, mime_type,
                               ignore_uncap=ignore_uncap, hier_delim=hier_delim,
                                                    strict_arrays=strict_arrays)

        self.tmp_dir = tmp_dir
        self.tmp_delete_on_close = tmp_delete_on_close
        self.parse_cookie = parse_cookie

    def get_tmp_delete_on_close(self):
        return self.__tmp_delete_on_close

    def set_tmp_delete_on_close(self, val):
        self.__tmp_delete_on_close = val
        self.stream_factory = get_stream_factory(self.tmp_dir,
                                                     self.__tmp_delete_on_close)

    tmp_delete_on_close = property(get_tmp_delete_on_close,
                                                        set_tmp_delete_on_close)

    def set_validator(self, validator):
        if validator == 'soft' or validator is self.SOFT_VALIDATION:
            self.validator = self.SOFT_VALIDATION
        elif validator is None:
            self.validator = None
        else:
            raise ValueError(validator)

    def create_in_document(self, ctx, in_string_encoding=None):
        assert ctx.transport.type.endswith('http'), \
            ("This protocol only works with an http transport, not %r, (in %r)"
                                          % (ctx.transport.type, ctx.transport))

        ctx.in_document = ctx.transport.req
        ctx.transport.request_encoding = in_string_encoding

    def decompose_incoming_envelope(self, ctx, message):
        assert message == SimpleDictDocument.REQUEST

        ctx.transport.itself.decompose_incoming_envelope(self, ctx, message)

        if self.parse_cookie:
            cookies = ctx.in_header_doc.get('cookie', None)
            if cookies is None:
                cookies = ctx.in_header_doc.get('Cookie', None)

            if cookies is not None:
                for cookie_string in cookies:
                    logger.debug("Loading cookie string %r", cookie_string)
                    cookie = _parse_cookie(cookie_string)
                    for k, v in cookie.items():
                        l = ctx.in_header_doc.get(k, [])
                        l.append(v)
                        ctx.in_header_doc[k] = l

        logger.debug('\theader : %r' % (ctx.in_header_doc))
        logger.debug('\tbody   : %r' % (ctx.in_body_doc))

    def deserialize(self, ctx, message):
        assert message in (self.REQUEST,)

        self.event_manager.fire_event('before_deserialize', ctx)

        if ctx.descriptor is None:
            raise ResourceNotFoundError(ctx.method_request_string)

        req_enc = getattr(ctx.transport, 'request_encoding', None)
        if req_enc is None:
            req_enc = ctx.in_protocol.default_string_encoding

        if ctx.descriptor.in_header is not None:
            # HttpRpc supports only one header class
            in_header_class = ctx.descriptor.in_header[0]
            ctx.in_header = self.simple_dict_to_object(ctx, ctx.in_header_doc,
                            in_header_class, self.validator, req_enc=req_enc)

        if ctx.descriptor.in_message is not None:
            ctx.in_object = self.simple_dict_to_object(ctx, ctx.in_body_doc,
                    ctx.descriptor.in_message, self.validator, req_enc=req_enc)

        self.event_manager.fire_event('after_deserialize', ctx)

    def serialize(self, ctx, message):
        retval = None

        assert message in (self.RESPONSE,)

        if ctx.out_document is not None:
            return

        if ctx.out_error is not None:
            ctx.transport.mime_type = 'text/plain'
            ctx.out_document = ctx.out_error.to_bytes_iterable(ctx.out_error)

        else:
            retval = self._handle_rpc(ctx)

        self.event_manager.fire_event('serialize', ctx)

        return retval

    @coroutine
    def _handle_rpc_nonempty(self, ctx):
        result_class = ctx.descriptor.out_message

        out_class = None
        out_object = None

        if ctx.descriptor.body_style is BODY_STYLE_WRAPPED:
            fti = result_class.get_flat_type_info(result_class)

            if len(fti) > 1 and not self.ignore_uncap:
                raise TypeError("HttpRpc protocol can only serialize "
                                "functions with a single return type.")

            if len(fti) == 1:
                out_class, = fti.values()
                out_object, = ctx.out_object

        else:
            out_class = result_class
            out_object, = ctx.out_object

        if out_class is not None:
            if issubclass(out_class, File) and not \
                        isinstance(out_object, (list, tuple, string_types)) \
                        and out_object.type is not None:
                ctx.transport.set_mime_type(str(out_object.type))

            ret = self.to_bytes_iterable(out_class, out_object)

            if not isinstance(ret, PushBase):
                ctx.out_document = ret

            else:
                ctx.transport.itself.set_out_document_push(ctx)
                while True:
                    sv = yield
                    ctx.out_document.send(sv)

    def _handle_rpc(self, ctx):
        retval = None

        # assign raw result to its wrapper, result_message
        if ctx.out_object is None or len(ctx.out_object) < 1:
            ctx.out_document = ['']

        else:
            retval = self._handle_rpc_nonempty(ctx)

        header_class = ctx.descriptor.out_header
        if header_class is not None:
            # HttpRpc supports only one header class
            header_class = header_class[0]

        # header
        if ctx.out_header is not None:
            out_header = ctx.out_header
            if isinstance(ctx.out_header, (list, tuple)):
                out_header = ctx.out_header[0]

            ctx.out_header_doc = self.object_to_simple_dict(header_class,
                out_header, subinst_eater=_header_to_bytes)

        return retval

    def create_out_string(self, ctx, out_string_encoding='utf8'):
        if ctx.out_string is not None:
            return

        ctx.out_string = ctx.out_document

    def boolean_from_bytes(self, cls, string):
        return string.lower() in ('true', '1', 'checked', 'on')

    def integer_from_bytes(self, cls, string):
        if string == '':
            return None

        return super(HttpRpc, self).integer_from_bytes(cls, string)


_fragment_pattern_re = re.compile('<([A-Za-z0-9_]+)>')
_full_pattern_re = re.compile('{([A-Za-z0-9_]+)}')


class HttpPattern(object):
    """Experimental. Stay away.

    :param address: Address pattern
    :param verb: HTTP Verb pattern
    :param host: HTTP "Host:" header pattern
    """

    @staticmethod
    def _compile_url_pattern(pattern):
        """where <> placeholders don't contain slashes."""

        if pattern is None:
            return None
        pattern = _fragment_pattern_re.sub(r'(?P<\1>[^/]*)', pattern)
        pattern = _full_pattern_re.sub(r'(?P<\1>[^/]*)', pattern)
        return re.compile(pattern)

    @staticmethod
    def _compile_host_pattern(pattern):
        """where <> placeholders don't contain dots."""

        if pattern is None:
            return None
        pattern = _fragment_pattern_re.sub(r'(?P<\1>[^\.]*)', pattern)
        pattern = _full_pattern_re.sub(r'(?P<\1>.*)', pattern)
        return re.compile(pattern)

    @staticmethod
    def _compile_verb_pattern(pattern):
        """where <> placeholders are same as {} ones."""

        if pattern is None:
            return None
        pattern = _fragment_pattern_re.sub(r'(?P<\1>.*)', pattern)
        pattern = _full_pattern_re.sub(r'(?P<\1>.*)', pattern)
        return re.compile(pattern)

    def __init__(self, address=None, verb=None, host=None, endpoint=None):
        self.address = address
        self.host = host
        self.verb = verb

        self.endpoint = endpoint
        if self.endpoint is not None:
            assert isinstance(self.endpoint, MethodDescriptor)

    def hello(self, descriptor):
        if self.address is None:
            self.address = descriptor.name

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, what):
        if what is not None and not what.startswith('/'):
            what = '/' + what

        self.__address = what
        self.address_re = self._compile_url_pattern(what)

    @property
    def host(self):
        return self.__host

    @host.setter
    def host(self, what):
        self.__host = what
        self.host_re = self._compile_host_pattern(what)

    @property
    def verb(self):
        return self.__verb

    @verb.setter
    def verb(self, what):
        self.__verb = what
        self.verb_re = self._compile_verb_pattern(what)

    def as_werkzeug_rule(self):
        from werkzeug.routing import Rule
        from spyne.util.invregexp import invregexp

        methods = None
        if self.verb is not None:
            methods = invregexp(self.verb)

        host = self.host
        if host is None:
            host = '<__ignored>'  # for some reason, this is necessary when
                                  # host_matching is enabled.

        return Rule(self.address, host=host, endpoint=self.endpoint.name,
                                                                methods=methods)

    def __repr__(self):
        return "HttpPattern(address=%r, host=%r, verb=%r, endpoint=%r)" % (
                    self.address, self.host, self.verb,
                    None if self.endpoint is None else self.endpoint.name)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""
A barely functional Spyne class serializer. If you're using this as part of
anything serious, you're insane.

TODO:
  - Customizations are not serialized.
"""

import logging
logger = logging.getLogger(__name__)

import spyne

from datetime import datetime
from itertools import chain
from collections import defaultdict

from spyne.model import SimpleModel
from spyne.model.complex import XmlModifier
from spyne.model.complex import ComplexModelBase


def gen_fn_from_tns(tns):
    return tns \
        .replace('http://', '') \
        .replace('https://', '') \
        .replace('/', '') \
        .replace('.', '_') \
        .replace(':', '_') \
        .replace('#', '') \
        .replace('-', '_')


class CodeGenerator(object):
    def __init__(self, fn_tns_mapper=gen_fn_from_tns):
        self.imports = set()
        self.classes = set()
        self.pending = defaultdict(list)
        self.simples = set()
        self.fn_tns_mapper = fn_tns_mapper

    def gen_modifier(self, t):
        return '%s(%s)' % (t.__name__, self.gen_dispatch(t.type))

    def gen_simple(self, t):
        return t.__name__

    def gen_complex(self, t):
        retval = []
        retval.append("""

class %s(_ComplexBase):
    _type_info = [""" % (t.get_type_name()))

        for k,v in t._type_info.items():
            if not issubclass(v, ComplexModelBase) or \
                        v.get_namespace() != self.tns or \
                            v in self.classes or \
                                getattr(v, '__orig__', None) in self.classes:
                retval.append("        ('%s', %s)," % (k, self.gen_dispatch(v)))
            else:
                self.pending[v.get_type_name()].append((k, t.get_type_name()))

        retval.append("    ]")

        self.classes.add(t)

        for k,orig_t in self.pending[t.get_type_name()]:
            retval.append('%s._type_info["%s"] = %s' % (orig_t, k, t.get_type_name()))

        return retval

    def gen_dispatch(self, t):
        if issubclass(t, XmlModifier):
            return self.gen_modifier(t)

        if issubclass(t, SimpleModel):
            return self.gen_simple(t)

        if t.get_namespace() == self.tns:
            return t.get_type_name()

        i = self.fn_tns_mapper(t.get_namespace())
        self.imports.add(i)
        return "%s.%s" % (i, t.get_type_name())

    def genpy(self, tns, s):
        self.tns = tns

        retval = [u"""# encoding: utf8

# Automatically generated by Spyne %s at %s.
# Modify at your own risk.

from spyne.model import *
"""  % (spyne.__version__, datetime.now().replace(microsecond=0).isoformat(' ')),
"", # imports
"""

class _ComplexBase(ComplexModelBase):
    __namespace__ = '%s'
    __metaclass__ = ComplexModelMeta""" % tns
]

        for n, t in s.types.items():
            if issubclass(t, ComplexModelBase):
                retval.extend(self.gen_complex(t))
            else:
                retval.append('%s = %s' % (n, self.gen_dispatch(t)))
                self.simples.add(n)

        for i in self.imports:
            retval.insert(1, "import %s" % i)

        retval.append("")
        retval.append("")

        retval.append('__all__ = [')
        for c in sorted(chain([c.get_type_name() for c in self.classes],
                                                                self.simples)):
            retval.append("    '%s',"  % c)
        retval.append(']')
        retval.append("")

        return '\n'.join(retval)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import getpass
import inspect
import traceback
import smtplib

from socket import gethostname
from subprocess import Popen, PIPE

from email.utils import COMMASPACE, formatdate
from email.mime.text import MIMEText

from spyne.util import six


def email_exception(exception_address, message="", bcc=None):
    # http://stackoverflow.com/questions/1095601/find-module-name-of-the-originating-exception-in-python
    frm = inspect.trace()[-1]
    mod = inspect.getmodule(frm[0])
    module_name = mod.__name__ if mod else frm[1]

    sender = 'robot@spyne.io'
    recipients = [exception_address]
    if bcc is not None:
        recipients.extend(bcc)

    error_str = ("%s\n\n%s" % (message, traceback.format_exc()))
    msg = MIMEText(error_str.encode('utf8'), 'plain', 'utf8')
    msg['To'] = exception_address
    msg['From'] = 'Spyne <robot@spyne.io>'
    msg['Date'] = formatdate()
    msg['Subject'] = "(%s@%s) %s" % (getpass.getuser(), gethostname(), module_name)

    try:
        smtp_object = smtplib.SMTP('localhost')
        smtp_object.sendmail(sender, recipients, msg.as_string())
        logger.error("Error email sent")

    except Exception as e:
        logger.error("Error: unable to send email")
        logger.exception(e)


def email_text_smtp(addresses, sender=None, subject='', message="",
                                                     host='localhost', port=25):
    if sender is None:
        sender = 'Spyne <robot@spyne.io>'

    exc = traceback.format_exc()
    if exc is not None:
        message = (u"%s\n\n%s" % (message, exc))
    msg = MIMEText(message.encode('utf8'), 'plain', 'utf8')
    msg['To'] = COMMASPACE.join(addresses)
    msg['From'] = sender
    msg['Date'] = formatdate()
    msg['Subject'] = subject

    smtp_object = smtplib.SMTP(host, port)
    if six.PY2:
        smtp_object.sendmail(sender, addresses, msg.as_string())
    else:
        smtp_object.sendmail(sender, addresses, msg.as_bytes())
    logger.info("Text email sent to: %r.", addresses)


def email_text(addresses, sender=None, subject='', message="", bcc=None):
    if sender is None:
        sender = 'Spyne <robot@spyne.io>'

    exc = traceback.format_exc()
    if exc is not None:
        message = (u"%s\n\n%s" % (message, exc))
    msg = MIMEText(message.encode('utf8'), 'plain', 'utf8')
    msg['To'] = COMMASPACE.join(addresses)
    msg['From'] = sender
    msg['Date'] = formatdate()
    msg['Subject'] = subject

    cmd = ["/usr/sbin/sendmail", "-oi", '--']
    cmd.extend(addresses)
    if bcc is not None:
        cmd.extend(bcc)

    p = Popen(cmd, stdin=PIPE)
    if six.PY2:
        p.communicate(msg.as_string())
    else:
        p.communicate(msg.as_bytes())

    logger.info("Text email sent to: %r.", addresses)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import re
from spyne import M, Boolean, DateTime, Date, Time, ComplexModel, \
    ValidationError
from spyne.protocol import InProtocolBase


class SegmentBase(object):
    @classmethod
    def from_string(cls, s):
        match = cls._SEGMENT_RE.match(s)
        if match is None:
            raise ValidationError(s)
        start_incl, start_str, end_str, end_incl = match.groups()

        print()
        print(start_incl, start_str, end_str, end_incl)

        start_incl = (start_incl == '[')
        start = InProtocolBase().from_unicode(
                                            cls._type_info['start'], start_str)
        end = InProtocolBase().from_unicode(cls._type_info['start'], end_str)
        end_incl = (end_incl == ']')

        print(start_incl, start, end, end_incl)

        return cls(start_inclusive=start_incl, start=start, end=end,
                                                         end_inclusive=end_incl)

    def to_string(self):
        return '[%s,%s]' % (self.start.isoformat(), self.end.isoformat())


class DateTimeSegment(ComplexModel, SegmentBase):
    _SEGMENT_RE = re.compile(
        u"([\\[\\]])"
        u"([0-9:\\.T-]+)"
        u","
        u"([0-9:\\.T-]+)"
        u"([\\[\\]])", re.DEBUG | re.UNICODE)

    _type_info = [
        ('start_inclusive', M(Boolean(default=True))),
        ('start', M(DateTime)),
        ('end', M(DateTime)),
        ('end_inclusive', M(Boolean(default=True))),
    ]



class DateSegment(ComplexModel, SegmentBase):
    _SEGMENT_RE = re.compile(
        u"([\\[\\]])"
        u"([0-9-]+)"
        u","
        u"([0-9-]+)"
        u"([\\[\\]])", re.DEBUG | re.UNICODE)

    _type_info = [
        ('start_inclusive', M(Boolean(default=True))),
        ('start', M(Date)),
        ('end', M(Date)),
        ('end_inclusive', M(Boolean(default=True))),
    ]


class TimeSegment(ComplexModel, SegmentBase):
    _SEGMENT_RE = re.compile(
        u"([\\[\\]])"
        u"([0-9:\\.]+)"
        u","
        u"([0-9:\\.]+)"
        u"([\\[\\]])", re.DEBUG | re.UNICODE)

    _type_info = [
        ('start_inclusive', M(Boolean(default=True))),
        ('start', M(Time)),
        ('end', M(Time)),
        ('end_inclusive', M(Boolean(default=True))),
    ]




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.interface.xml_schema.model`` module contains type-specific logic
for schema generation."""


import logging
logger = logging.getLogger(__name__)

from decimal import Decimal as D
from collections import deque, defaultdict

from lxml import etree

from spyne.model import ModelBase, XmlAttribute, AnyXml, Unicode, XmlData, \
    Decimal, Integer
from spyne.const.xml import NS_XSD, XSD
from spyne.util import memoize
from spyne.util.cdict import cdict
from spyne.util.etreeconv import dict_to_etree
from spyne.util.six import string_types
from spyne.protocol.xml import XmlDocument
_prot = XmlDocument()

# In Xml Schema, some customizations do not need a class to be extended -- they
# are specified in-line in the parent class definition, like nullable or
# min_occurs. The dict below contains list of parameters that do warrant a
# proper subclass definition for each type. This must be updated as the Xml
# Schema implementation makes progress.
ATTR_NAMES = cdict({
    ModelBase: set(['values']),
    Decimal: set(['pattern', 'gt', 'ge', 'lt', 'le', 'values', 'total_digits',
                                                            'fraction_digits']),
    Integer: set(['pattern', 'gt', 'ge', 'lt', 'le', 'values', 'total_digits']),
    Unicode: set(['values', 'min_len', 'max_len', 'pattern']),
})

def xml_attribute_add(cls, name, element, document):
    element.set('name', name)
    element.set('type', cls.type.get_type_name_ns(document.interface))

    if cls._use is not None:
        element.set('use', cls._use)

    d = cls.type.Attributes.default

    if d is not None:
        element.set('default', _prot.to_unicode(cls.type, d))


def _check_extension_attrs(cls):
    """Make sure only customizations that need a restriction tag generate one"""

    extends = cls.__extends__

    eattrs = extends.Attributes
    cattrs = cls.Attributes

    ckeys = set([k for k in vars(cls.Attributes) if not k.startswith('_')])
    ekeys = set([k for k in vars(extends.Attributes) if not k.startswith('_')])

    # get the attributes different from the parent class
    diff = set()
    for k in (ckeys | ekeys):
        if getattr(eattrs, k, None) != getattr(cattrs, k, None):
            diff.add(k)

    # compare them with what comes from ATTR_NAMES
    attr_names = ATTR_NAMES[cls]
    retval = None
    while extends is not None:
        retval = extends
        if len(diff & attr_names) > 0:
            return extends
        extends = extends.__extends__

    return retval

# noinspection PyDefaultArgument
def simple_get_restriction_tag(document, cls):
    extends = _check_extension_attrs(cls)
    if extends is None:
        return

    simple_type = etree.Element(XSD('simpleType'))

    simple_type.set('name', cls.get_type_name())
    document.add_simple_type(cls, simple_type)

    restriction = etree.SubElement(simple_type, XSD('restriction'))
    restriction.set('base', extends.get_type_name_ns(document.interface))

    for v in cls.Attributes.values:
        enumeration = etree.SubElement(restriction, XSD('enumeration'))
        enumeration.set('value', XmlDocument().to_unicode(cls, v))

    return restriction


def simple_add(document, cls, tags):
    if not cls.is_default(cls):
        document.get_restriction_tag(cls)

def byte_array_add(document, cls, tags):
    simple_add(document, cls, tags)


def complex_add(document, cls, tags):
    complex_type = etree.Element(XSD('complexType'))
    complex_type.set('name', cls.get_type_name())

    doc_text = cls.get_documentation()
    if doc_text or cls.Annotations.appinfo is not None:
        annotation = etree.SubElement(complex_type, XSD('annotation'))
        if doc_text:
            doc = etree.SubElement(annotation, XSD('documentation'))
            doc.text = doc_text

        _ai = cls.Annotations.appinfo
        if _ai is not None:
            appinfo = etree.SubElement(annotation, XSD('appinfo'))
            if isinstance(_ai, dict):
                dict_to_etree(_ai, appinfo)

            elif isinstance(_ai, string_types):
                appinfo.text = _ai

            elif isinstance(_ai, etree._Element):
                appinfo.append(_ai)

            else:
                from spyne.util.xml import get_object_as_xml

                appinfo.append(get_object_as_xml(_ai))

    sequence_parent = complex_type
    extends = getattr(cls, '__extends__', None)

    type_info = cls._type_info
    if extends is not None:
        if (extends.get_type_name() == cls.get_type_name() and
                               extends.get_namespace() == cls.get_namespace()):
            raise Exception("%r can't extend %r because they are both '{%s}%s'"
                    % (cls, extends, cls.get_namespace(), cls.get_type_name()))

        if extends.Attributes.exc_interface:
            # If the parent class is private, it won't be in the schema, so we
            # need to act as if its attributes are part of cls as well.
            type_info = cls.get_simple_type_info(cls)

        else:
            complex_content = etree.SubElement(complex_type,
                                                          XSD('complexContent'))
            extension = etree.SubElement(complex_content, XSD('extension'))
            extension.set('base', extends.get_type_name_ns(document.interface))
            sequence_parent = extension

    if cls.Attributes._xml_tag_body_as is not None:
        for xtba_key, xtba_type in cls.Attributes._xml_tag_body_as:
            _sc = etree.SubElement(sequence_parent, XSD('simpleContent'))
            xtba_ext = etree.SubElement(_sc, XSD('extension'))
            xtba_ext.attrib['base'] = xtba_type.type.get_type_name_ns(
                                                             document.interface)

    sequence = etree.Element(XSD('sequence'))

    deferred = deque()
    choice_tags = defaultdict(lambda: etree.Element(XSD('choice')))

    for k, v in type_info.items():
        assert isinstance(k, string_types)
        assert issubclass(v, ModelBase)

        a = v.Attributes
        if a.exc_interface:
            continue

        if issubclass(v, XmlData):
            continue

        if issubclass(v, XmlAttribute):
            deferred.append((k,v))
            continue

        document.add(v, tags)

        name = a.sub_name
        if name is None:
            name = k
        #ns = a.sub_ns
        #if ns is not None:
        #    name = "{%s}%s" % (ns, name)

        type_name_ns = v.get_type_name_ns(document.interface)
        if v.__extends__ is not None and v.__orig__ is not None and \
                                              _check_extension_attrs(v) is None:
            type_name_ns = v.__orig__.get_type_name_ns(document.interface)

        member = etree.Element(a.schema_tag)
        if a.schema_tag == XSD('element'):
            member.set('name', name)
            member.set('type', type_name_ns)

        elif a.schema_tag == XSD('any') and issubclass(v, AnyXml):
            if a.namespace is not None:
                member.set('namespace', a.namespace)
            if a.process_contents is not None:
                member.set('processContents', a.process_contents)

        else:
            raise ValueError("Unhandled schema_tag / type combination. %r %r"
                    % (v, a.schema_tag))

        if a.min_occurs != 1:  # 1 is the xml schema default
            member.set('minOccurs', str(a.min_occurs))

        if a.max_occurs != 1:  # 1 is the xml schema default
            val = a.max_occurs
            if val in (D('inf'), float('inf')):
                val = 'unbounded'
            else:
                val = str(val)

            member.set('maxOccurs', val)

        if a.default is not None:
            member.set('default', _prot.to_unicode(v, a.default))

        if bool(a.nillable) != False: # False is the xml schema default
            member.set('nillable', 'true')

        v_doc_text = v.get_documentation()
        if v_doc_text:
            # Doesn't support multi-language documentation
            annotation = etree.SubElement(member, XSD('annotation'))
            doc = etree.SubElement(annotation, XSD('documentation'))
            doc.text = v_doc_text

        if a.xml_choice_group is None:
            sequence.append(member)
        else:
            choice_tags[a.xml_choice_group].append(member)

    sequence.extend(choice_tags.values())

    if len(sequence) > 0:
        sequence_parent.append(sequence)

    _ext_elements = dict()
    for k,v in deferred:
        attribute = etree.Element(XSD('attribute'))
        xml_attribute_add(v, k, attribute, document)

        if cls.Attributes._xml_tag_body_as is None:
            sequence_parent.append(attribute)
        else:
            xtba_ext.append(attribute)

    document.add_complex_type(cls, complex_type)

    # simple node
    complex_type_name = cls.Attributes.sub_name or cls.get_type_name()
    element = etree.Element(XSD('element'))
    element.set('name', complex_type_name)
    element.set('type', cls.get_type_name_ns(document.interface))

    document.add_element(cls, element)


def enum_add(document, cls, tags):
    simple_type = etree.Element(XSD('simpleType'))
    simple_type.set('name', cls.get_type_name())

    restriction = etree.SubElement(simple_type, XSD('restriction'))
    restriction.set('base', '%s:string' %
                               document.interface.get_namespace_prefix(NS_XSD))

    for v in cls.__values__:
        enumeration = etree.SubElement(restriction, XSD('enumeration'))
        enumeration.set('value', v)

    document.add_simple_type(cls, simple_type)

fault_add = complex_add


def unicode_get_restriction_tag(document, cls):
    restriction = simple_get_restriction_tag(document, cls)
    if restriction is None:
        return

    # length
    if cls.Attributes.min_len == cls.Attributes.max_len:
        length = etree.SubElement(restriction, XSD('length'))
        length.set('value', str(cls.Attributes.min_len))

    else:
        if cls.Attributes.min_len != Unicode.Attributes.min_len:
            min_l = etree.SubElement(restriction, XSD('minLength'))
            min_l.set('value', str(cls.Attributes.min_len))

        if cls.Attributes.max_len != Unicode.Attributes.max_len:
            max_l = etree.SubElement(restriction, XSD('maxLength'))
            max_l.set('value', str(cls.Attributes.max_len))

    # pattern
    if cls.Attributes.pattern != Unicode.Attributes.pattern:
        pattern = etree.SubElement(restriction, XSD('pattern'))
        pattern.set('value', cls.Attributes.pattern)

    return restriction


prot = XmlDocument()


def Tget_range_restriction_tag(T):
    """The get_range_restriction template function. Takes a primitive, returns
    a function that generates range restriction tags.
    """

    from spyne.model.primitive import Decimal
    from spyne.model.primitive import Integer

    if issubclass(T, Decimal):
        def _get_float_restrictions(prot, restriction, cls):
            if cls.Attributes.fraction_digits != T.Attributes.fraction_digits:
                elt = etree.SubElement(restriction, XSD('fractionDigits'))
                elt.set('value', prot.to_unicode(cls,
                                                cls.Attributes.fraction_digits))

        def _get_integer_restrictions(prot, restriction, cls):
            if cls.Attributes.total_digits != T.Attributes.total_digits:
                elt = etree.SubElement(restriction, XSD('totalDigits'))
                elt.set('value', prot.to_unicode(cls,
                                                   cls.Attributes.total_digits))

        if issubclass(T, Integer):
            def _get_additional_restrictions(prot, restriction, cls):
                _get_integer_restrictions(prot, restriction, cls)

        else:
            def _get_additional_restrictions(prot, restriction, cls):
                _get_integer_restrictions(prot, restriction, cls)
                _get_float_restrictions(prot, restriction, cls)

    else:
        def _get_additional_restrictions(prot, restriction, cls):
            pass

    def _get_range_restriction_tag(document, cls):
        restriction = simple_get_restriction_tag(document, cls)
        if restriction is None:
            return

        if cls.Attributes.gt != T.Attributes.gt:
            elt = etree.SubElement(restriction, XSD('minExclusive'))
            elt.set('value', prot.to_unicode(cls, cls.Attributes.gt))

        if cls.Attributes.ge != T.Attributes.ge:
            elt = etree.SubElement(restriction, XSD('minInclusive'))
            elt.set('value', prot.to_unicode(cls, cls.Attributes.ge))

        if cls.Attributes.lt != T.Attributes.lt:
            elt = etree.SubElement(restriction, XSD('maxExclusive'))
            elt.set('value', prot.to_unicode(cls, cls.Attributes.lt))

        if cls.Attributes.le != T.Attributes.le:
            elt = etree.SubElement(restriction, XSD('maxInclusive'))
            elt.set('value', prot.to_unicode(cls, cls.Attributes.le))

        if cls.Attributes.pattern != T.Attributes.pattern:
            elt = etree.SubElement(restriction, XSD('pattern'))
            elt.set('value', cls.Attributes.pattern)

        _get_additional_restrictions(prot, restriction, cls)

        return restriction

    return _get_range_restriction_tag




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


from __future__ import absolute_import, print_function

import logging
logger = logging.getLogger(__name__)

import sqlalchemy

try:
    import simplejson as json
except ImportError:
    import json

from os.path import isabs
from inspect import isclass

from sqlalchemy import event
from sqlalchemy.schema import Column
from sqlalchemy.schema import Index
from sqlalchemy.schema import Table
from sqlalchemy.schema import ForeignKey
from sqlalchemy.orm import _mapper_registry

from sqlalchemy.dialects.postgresql import FLOAT
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION
from sqlalchemy.dialects.postgresql.base import PGUuid, PGInet

from sqlalchemy.orm import relationship
from sqlalchemy.orm import mapper
from sqlalchemy.ext.associationproxy import association_proxy

from spyne.store.relational.simple import PGLTree
from spyne.store.relational.document import PGXml, PGObjectXml, PGObjectJson, \
    PGFileJson, PGJsonB, PGHtml, PGJson
from spyne.store.relational.spatial import PGGeometry

# internal types
from spyne.model.enum import EnumBase
from spyne.model.complex import XmlModifier

# Config types
from spyne.model import xml as c_xml
from spyne.model import json as c_json
from spyne.model import jsonb as c_jsonb
from spyne.model import table as c_table
from spyne.model import msgpack as c_msgpack
from spyne.model.binary import HybridFileStore

# public types
from spyne.model import SimpleModel, Enum, Array, ComplexModelBase, \
    Any, AnyDict, AnyXml, AnyHtml, \
    Date, Time, DateTime, Duration, \
    ByteArray, String, Unicode, Uuid, Boolean, \
    Point, Line, Polygon, MultiPoint, MultiLine, MultiPolygon, \
    Float, Double, Decimal, \
    Integer, Integer8, Integer16, Integer32, Integer64, \
    UnsignedInteger, UnsignedInteger8, UnsignedInteger16, UnsignedInteger32, \
                                                            UnsignedInteger64, \
    Ipv6Address, Ipv4Address, IpAddress, \
    File, Ltree

from spyne.util import sanitize_args


# Inheritance type constants.
class _SINGLE:
    pass

class _JOINED:
    pass


_sq2sp_type_map = {
    # we map float => double because sqla doesn't
    # distinguish between floats and doubles.
    sqlalchemy.Float: Double,
    sqlalchemy.FLOAT: Double,

    sqlalchemy.Numeric: Decimal,
    sqlalchemy.NUMERIC: Decimal,

    sqlalchemy.BigInteger: Integer64,
    sqlalchemy.BIGINT: Integer64,

    sqlalchemy.Integer: Integer32,
    sqlalchemy.INTEGER: Integer32,

    sqlalchemy.SmallInteger: Integer16,
    sqlalchemy.SMALLINT: Integer16,

    sqlalchemy.Binary: ByteArray,
    sqlalchemy.LargeBinary: ByteArray,

    sqlalchemy.Boolean: Boolean,
    sqlalchemy.BOOLEAN: Boolean,

    sqlalchemy.DateTime: DateTime,
    sqlalchemy.TIMESTAMP: DateTime,
    sqlalchemy.dialects.postgresql.base.TIMESTAMP: DateTime,
    sqlalchemy.DATETIME: DateTime,
    sqlalchemy.dialects.postgresql.base.INTERVAL: Duration,

    sqlalchemy.Date: Date,
    sqlalchemy.DATE: Date,

    sqlalchemy.Time: Time,
    sqlalchemy.TIME: Time,

    PGUuid: Uuid,
    PGLTree: Ltree,
    PGInet: IpAddress,
}


# this needs to be called whenever a new column is instantiated.
def _sp_attrs_to_sqla_constraints(cls, subcls, col_kwargs=None, col=None):
    # cls is the parent class of v
    if subcls.Attributes.nullable == False and cls.__extends__ is None:
        if col is None:
            col_kwargs['nullable'] = False
        else:
            col.nullable = False

    if subcls.Attributes.db_default is not None:
        if col is None:
            col_kwargs['default'] = subcls.Attributes.db_default
        else:
            col.default = subcls.Attributes.db_default


def _get_sqlalchemy_type(cls):
    db_type = cls.Attributes.db_type
    if db_type is not None:
        return db_type

    # must be above Unicode, because Ltree is Unicode's subclass
    if issubclass(cls, Ltree):
        return PGLTree

    # must be above Unicode, because Ip*Address is Unicode's subclass
    if issubclass(cls, (IpAddress, Ipv4Address, Ipv6Address)):
        return PGInet

    # must be above Unicode, because Uuid is Unicode's subclass
    if issubclass(cls, Uuid):
        return PGUuid(as_uuid=True)

    # must be above Unicode, because Point is Unicode's subclass
    if issubclass(cls, Point):
        return PGGeometry("POINT", dimension=cls.Attributes.dim)

    # must be above Unicode, because Line is Unicode's subclass
    if issubclass(cls, Line):
        return PGGeometry("LINESTRING", dimension=cls.Attributes.dim)

    # must be above Unicode, because Polygon is Unicode's subclass
    if issubclass(cls, Polygon):
        return PGGeometry("POLYGON", dimension=cls.Attributes.dim)

    # must be above Unicode, because MultiPoint is Unicode's subclass
    if issubclass(cls, MultiPoint):
        return PGGeometry("MULTIPOINT", dimension=cls.Attributes.dim)

    # must be above Unicode, because MultiLine is Unicode's subclass
    if issubclass(cls, MultiLine):
        return PGGeometry("MULTILINESTRING", dimension=cls.Attributes.dim)

    # must be above Unicode, because MultiPolygon is Unicode's subclass
    if issubclass(cls, MultiPolygon):
        return PGGeometry("MULTIPOLYGON", dimension=cls.Attributes.dim)

    # must be above Unicode, because String is Unicode's subclass
    if issubclass(cls, String):
        if cls.Attributes.max_len == String.Attributes.max_len: # Default is arbitrary-length
            return sqlalchemy.Text
        else:
            return sqlalchemy.String(cls.Attributes.max_len)

    if issubclass(cls, Unicode):
        if cls.Attributes.max_len == Unicode.Attributes.max_len: # Default is arbitrary-length
            return sqlalchemy.UnicodeText
        else:
            return sqlalchemy.Unicode(cls.Attributes.max_len)

    if issubclass(cls, EnumBase):
        return sqlalchemy.Enum(*cls.__values__, name=cls.__type_name__)

    if issubclass(cls, AnyXml):
        return PGXml

    if issubclass(cls, AnyHtml):
        return PGHtml

    if issubclass(cls, (Any, AnyDict)):
        sa = cls.Attributes.store_as
        if sa is None:
            return None
        if isinstance(sa, c_json):
            return PGJson
        if isinstance(sa, c_jsonb):
            return PGJsonB
        raise NotImplementedError(dict(cls=cls, store_as=sa))

    if issubclass(cls, ByteArray):
        return sqlalchemy.LargeBinary

    if issubclass(cls, (Integer64, UnsignedInteger64)):
        return sqlalchemy.BigInteger

    if issubclass(cls, (Integer32, UnsignedInteger32)):
        return sqlalchemy.Integer

    if issubclass(cls, (Integer16, UnsignedInteger16)):
        return sqlalchemy.SmallInteger

    if issubclass(cls, (Integer8, UnsignedInteger8)):
        return sqlalchemy.SmallInteger

    if issubclass(cls, Float):
        return FLOAT

    if issubclass(cls, Double):
        return DOUBLE_PRECISION

    if issubclass(cls, (Integer, UnsignedInteger)):
        return sqlalchemy.DECIMAL

    if issubclass(cls, Decimal):
        return sqlalchemy.DECIMAL

    if issubclass(cls, Boolean):
        if cls.Attributes.store_as is bool:
            return sqlalchemy.Boolean
        if cls.Attributes.store_as is int:
            return sqlalchemy.SmallInteger

        raise ValueError("Boolean.store_as has invalid value %r" %
                                                        cls.Attributes.store_as)

    if issubclass(cls, Date):
        return sqlalchemy.Date

    if issubclass(cls, DateTime):
        if cls.Attributes.timezone is None:
            if cls.Attributes.as_timezone is None:
                return sqlalchemy.DateTime(timezone=True)
            else:
                return sqlalchemy.DateTime(timezone=False)
        else:
            return sqlalchemy.DateTime(timezone=cls.Attributes.timezone)

    if issubclass(cls, Time):
        return sqlalchemy.Time

    if issubclass(cls, Duration):
        return sqlalchemy.dialects.postgresql.base.INTERVAL

    if issubclass(cls, XmlModifier):
        retval = _get_sqlalchemy_type(cls.type)
        return retval


def _get_col_o2o(parent, subname, subcls, fk_col_name, deferrable=None,
                                  initially=None, ondelete=None, onupdate=None):
    """Gets key and child type and returns a column that points to the primary
    key of the child.
    """

    assert subcls.Attributes.table_name is not None, \
                                                "%r has no table name." % subcls

    col_args, col_kwargs = sanitize_args(subcls.Attributes.sqla_column_args)
    _sp_attrs_to_sqla_constraints(parent, subcls, col_kwargs)

    # get pkeys from child class
    pk_column, = get_pk_columns(subcls)  # FIXME: Support multi-col keys

    pk_key, pk_spyne_type = pk_column
    pk_sqla_type = _get_sqlalchemy_type(pk_spyne_type)

    # generate a fk to it from the current object (cls)
    if 'name' in col_kwargs:
        colname = col_kwargs.pop('name')
    else:
        colname = subname

    if fk_col_name is None:
        fk_col_name = colname + "_" + pk_key

    assert fk_col_name != colname, \
        "The column name for the foreign key must be different from the " \
        "column name for the object itself."

    fk = ForeignKey(
        '%s.%s' % (subcls.Attributes.table_name, pk_key),
        use_alter=True,
        name='%s_%s_fkey' % (subcls.Attributes.table_name, fk_col_name),
        deferrable=deferrable, initially=initially,
        ondelete=ondelete, onupdate=onupdate,
    )

    return Column(fk_col_name, pk_sqla_type, fk, **col_kwargs)


def _get_col_o2m(cls, fk_col_name, deferrable=None, initially=None,
                                                  ondelete=None, onupdate=None):
    """Gets the parent class and returns a column that points to the primary key
    of the parent.
    """

    assert cls.Attributes.table_name is not None, "%r has no table name." % cls
    col_args, col_kwargs = sanitize_args(cls.Attributes.sqla_column_args)

    # get pkeys from current class
    pk_column, = get_pk_columns(cls) # FIXME: Support multi-col keys

    pk_key, pk_spyne_type = pk_column
    pk_sqla_type = _get_sqlalchemy_type(pk_spyne_type)

    # generate a fk from child to the current class
    if fk_col_name is None:
        fk_col_name = '_'.join([cls.Attributes.table_name, pk_key])

    # we jump through all these hoops because we must instantiate the Column
    # only after we're sure that it doesn't already exist and also because
    # tinkering with functors is always fun :)
    yield [(fk_col_name, pk_sqla_type)]

    fk = ForeignKey('%s.%s' % (cls.Attributes.table_name, pk_key),
                                     deferrable=deferrable, initially=initially,
                                           ondelete=ondelete, onupdate=onupdate)
    col = Column(fk_col_name, pk_sqla_type, fk, **col_kwargs)

    yield col


def _get_cols_m2m(cls, k, child, fk_left_col_name, fk_right_col_name,
                  fk_left_deferrable, fk_left_initially,
                  fk_right_deferrable, fk_right_initially,
                  fk_left_ondelete, fk_left_onupdate,
                  fk_right_ondelete, fk_right_onupdate):
    """Gets the parent and child classes and returns foreign keys to both
    tables. These columns can be used to create a relation table."""

    col_info, left_col = _get_col_o2m(cls, fk_left_col_name,
                     ondelete=fk_left_ondelete, onupdate=fk_left_onupdate,
                     deferrable=fk_left_deferrable, initially=fk_left_initially)
    right_col = _get_col_o2o(cls, k, child, fk_right_col_name,
                   ondelete=fk_right_ondelete, onupdate=fk_right_onupdate,
                   deferrable=fk_right_deferrable, initially=fk_right_initially)
    left_col.primary_key = right_col.primary_key = True
    return left_col, right_col


class _FakeTable(object):
    def __init__(self, name):
        self.name = name
        self.c = {}
        self.columns = []
        self.indexes = []

    def append_column(self, col):
        self.columns.append(col)
        self.c[col.name] = col


def _gen_index_info(table, col, k, v):
    """
    :param table: sqla table
    :param col: sqla col
    :param k: field name (not necessarily == k)
    :param v: spyne type
    """

    unique = v.Attributes.unique
    index = v.Attributes.index
    if unique and not index:
        index = True

    try:
        index_name, index_method = index

    except (TypeError, ValueError):
        index_name = "%s_%s%s" % (table.name, k, '_unique' if unique else '')
        index_method = index

    if index in (False, None):
        return

    if index is True:
        index_args = (index_name, col), dict(unique=unique)
    else:
        index_args = (index_name, col), dict(unique=unique,
                                                  postgresql_using=index_method)

    if isinstance(table, _FakeTable):
        table.indexes.append(index_args)

    else:
        indexes = dict([(idx.name, idx) for idx in col.table.indexes])
        existing_idx = indexes.get(index_name, None)
        if existing_idx is None:
            Index(*index_args[0], **index_args[1])

        else:
            assert existing_idx.unique == unique
            assert existing_idx.kwargs.get('postgresql_using') == index_method


def _check_inheritance(cls, cls_bases):
    table_name = cls.Attributes.table_name

    inc = []
    inheritance = None
    base_class = getattr(cls, '__extends__', None)

    if base_class is None:
        for b in cls_bases:
            if getattr(b, '_type_info', None) is not None and b.__mixin__:
                base_class = b

    if base_class is not None:
        base_table_name = base_class.Attributes.table_name
        if base_table_name is not None:
            if base_table_name == table_name:
                inheritance = _SINGLE
            else:
                inheritance = _JOINED
                raise NotImplementedError("Joined table inheritance is not yet "
                                          "implemented.")

    # check whether the base classes are already mapped
    base_mapper = None
    if base_class is not None:
        base_mapper = base_class.Attributes.sqla_mapper

    if base_mapper is None:
        for b in cls_bases:
            bm = _mapper_registry.get(b, None)
            if bm is not None:
                assert base_mapper is None, "There can be only one base mapper."
                base_mapper = bm
                inheritance = _SINGLE

    return inheritance, base_class, base_mapper, inc


def _check_table(cls):
    table_name = cls.Attributes.table_name
    metadata = cls.Attributes.sqla_metadata

    # check whether the object already has a table
    table = None
    if table_name in metadata.tables:
        table = metadata.tables[table_name]
    else:
        # We need FakeTable because table_args can contain all sorts of stuff
        # that can require a fully-constructed table, and we don't have that
        # information here yet.
        table = _FakeTable(table_name)

    return table


def _add_simple_type(cls, props, table, subname, subcls, sqla_type):
    col_args, col_kwargs = sanitize_args(subcls.Attributes.sqla_column_args)
    _sp_attrs_to_sqla_constraints(cls, subcls, col_kwargs)

    mp = getattr(subcls.Attributes, 'mapper_property', None)

    if 'name' in col_kwargs:
        colname = col_kwargs.pop('name')
    else:
        colname = subname

    if not subcls.Attributes.exc_db:
        if colname in table.c:
            col = table.c[colname]

        else:
            col = Column(colname, sqla_type, *col_args, **col_kwargs)
            table.append_column(col)
            _gen_index_info(table, col, subname, subcls)

        if not subcls.Attributes.exc_mapper:
            props[subname] = col

    elif mp is not None:
        props[subname] = mp


def _gen_array_m2m(cls, props, subname, arrser, storage):
    """Generates a relational many-to-many array.

    :param cls: The class that owns the field
    :param props: SQLAlchemy Mapper properties
    :param subname: Field name
    :param arrser: Array serializer, ie the __orig__ of the class inside the
                   Array object
    :param storage: The storage configuration object passed to the store_as
    attribute.
    """

    metadata = cls.Attributes.sqla_metadata

    col_own, col_child = _get_cols_m2m(cls, subname, arrser,
                        storage.left, storage.right,
                        storage.fk_left_deferrable, storage.fk_left_initially,
                        storage.fk_right_deferrable, storage.fk_right_initially,
                        storage.fk_left_ondelete, storage.fk_left_onupdate,
                        storage.fk_right_ondelete, storage.fk_right_onupdate)

    storage.left = col_own.key
    storage.right = col_child.key

    # noinspection PySimplifyBooleanCheck because literal True means
    # "generate table name automatically" here
    if storage.multi is True:
        rel_table_name = '_'.join([cls.Attributes.table_name, subname])
    else:
        rel_table_name = storage.multi

    if rel_table_name in metadata.tables:
        rel_t = metadata.tables[rel_table_name]

        col_own_existing = rel_t.c.get(col_own.key, None)
        assert col_own_existing is not None
        if col_own_existing is not None:
            assert col_own.type.__class__ == col_own_existing.type.__class__

        col_child_existing = rel_t.c.get(col_child.key, None)
        if col_child_existing is None:
            rel_t.append_column(col_child)

        else:
            assert col_child.type.__class__ == col_child_existing.type.__class__

    else:
        rel_t = Table(rel_table_name, metadata, *(col_own, col_child))

    own_t = cls.Attributes.sqla_table

    rel_kwargs = dict(
        lazy=storage.lazy,
        backref=storage.backref,
        cascade=storage.cascade,
        order_by=storage.order_by,
        back_populates=storage.back_populates,
    )

    if storage.explicit_join:
        # Specify primaryjoin and secondaryjoin when requested.
        # There are special cases when sqlalchemy can't figure it out by itself.
        # this is where we help it when we can.
        # e.g.: http://sqlalchemy.readthedocs.org/en/rel_1_0/orm/join_conditions.html#self-referential-many-to-many-relationship

        assert own_t is not None and len(get_pk_columns(cls)) > 0

        # FIXME: support more than one pk
        (col_pk_key, _), = get_pk_columns(cls)
        col_pk = own_t.c[col_pk_key]

        rel_kwargs.update(dict(
            secondary=rel_t,
            primaryjoin=(col_pk == rel_t.c[col_own.key]),
            secondaryjoin=(col_pk == rel_t.c[col_child.key]),
        ))

        if storage.single_parent is not None:
            rel_kwargs['single_parent'] = storage.single_parent

        props[subname] = relationship(arrser, **rel_kwargs)

    else:
        rel_kwargs.update(dict(
            secondary=rel_t,
        ))

        if storage.single_parent is not None:
            rel_kwargs['single_parent'] = storage.single_parent

        props[subname] = relationship(arrser, **rel_kwargs)


def _gen_array_simple(cls, props, subname, arrser_cust, storage):
    """Generate an array of simple objects.

    :param cls: The class that owns this field
    :param props: SQLAlchemy Mapper properties
    :param subname: Field name
    :param arrser_cust: Array serializer, ie the class itself inside the Array
                        object
    :param storage: The storage configuration object passed to the store_as
    """

    table_name = cls.Attributes.table_name
    metadata = cls.Attributes.sqla_metadata

    # get left (fk) column info
    _gen_col = _get_col_o2m(cls, storage.left,
        ondelete=storage.fk_left_ondelete, onupdate=storage.fk_left_onupdate,
        deferrable=storage.fk_left_deferrable,
                                          initially=storage.fk_left_initially)

    col_info = next(_gen_col) # gets the column name
    # FIXME: Add support for multi-column primary keys.
    storage.left, child_left_col_type = col_info[0]
    child_left_col_name = storage.left

    # get right(data) column info
    child_right_col_type = _get_sqlalchemy_type(arrser_cust)
    child_right_col_name = storage.right  # this is the data column
    if child_right_col_name is None:
        child_right_col_name = subname

    # get table name
    child_table_name = arrser_cust.Attributes.table_name
    if child_table_name is None:
        child_table_name = '_'.join([table_name, subname])

    if child_table_name in metadata.tables:
        child_t = metadata.tables[child_table_name]

        # if we have the table, make sure have the right column (data column)
        assert child_right_col_type.__class__ is \
           child_t.c[child_right_col_name].type.__class__, "%s.%s: %r != %r" % \
                   (cls, child_right_col_name, child_right_col_type.__class__,
                               child_t.c[child_right_col_name].type.__class__)

        if child_left_col_name in child_t.c:
            assert child_left_col_type is \
                child_t.c[child_left_col_name].type.__class__, "%r != %r" % \
                   (child_left_col_type,
                               child_t.c[child_left_col_name].type.__class__)
        else:
            # Table exists but our own foreign key doesn't.
            child_left_col = next(_gen_col)
            _sp_attrs_to_sqla_constraints(cls, arrser_cust, col=child_left_col)
            child_t.append_column(child_left_col)

    else:
        # table does not exist, generate table
        child_right_col = Column(child_right_col_name, child_right_col_type)
        _sp_attrs_to_sqla_constraints(cls, arrser_cust, col=child_right_col)

        child_left_col = next(_gen_col)
        _sp_attrs_to_sqla_constraints(cls, arrser_cust, col=child_left_col)

        child_t = Table(child_table_name , metadata,
            Column('id', sqlalchemy.Integer, primary_key=True),
            child_left_col,
            child_right_col,
        )
        _gen_index_info(child_t, child_right_col, child_right_col_name,
                                                                    arrser_cust)

    # generate temporary class for association proxy
    cls_name = ''.join(x.capitalize() or '_' for x in
                                                child_table_name.split('_'))
                                                # generates camelcase class name.

    def _i(self, *args):
        setattr(self, child_right_col_name, args[0])

    cls_ = type("_" + cls_name, (object,), {'__init__': _i})
    mapper(cls_, child_t)
    props["_" + subname] = relationship(cls_)

    # generate association proxy
    setattr(cls, subname,
                         association_proxy("_" + subname, child_right_col_name))


def _gen_array_o2m(cls, props, subname, arrser, arrser_cust, storage):
    _gen_col = _get_col_o2m(cls, storage.right,
        ondelete=storage.fk_right_ondelete, onupdate=storage.fk_right_onupdate,
        deferrable=storage.fk_right_deferrable,
                                           initially=storage.fk_right_initially)

    col_info = next(_gen_col)  # gets the column name
    storage.right, col_type = col_info[0]  # FIXME: Add support for multi-column primary keys.

    assert storage.left is None, \
        "'left' is ignored in one-to-many relationships " \
        "with complex types (because they already have a " \
        "table). You probably meant to use 'right'."

    child_t = arrser.__table__

    if storage.right in child_t.c:
        # TODO: This branch MUST be tested.
        new_col_type = child_t.c[storage.right].type.__class__
        assert col_type is child_t.c[storage.right].type.__class__, \
                "Existing column type %r disagrees with new column type %r" % \
                                                        (col_type, new_col_type)

        # if the column is already there, the decision about whether
        # it should be in child's mapper or not should also have been
        # made.
        #
        # so, not adding the child column to to child mapper
        # here.
        col = child_t.c[storage.right]

    else:
        col = next(_gen_col)

        _sp_attrs_to_sqla_constraints(cls, arrser_cust, col=col)

        child_t.append_column(col)
        arrser.__mapper__.add_property(col.name, col)


    rel_kwargs = dict(
        lazy=storage.lazy,
        backref=storage.backref,
        cascade=storage.cascade,
        order_by=storage.order_by,
        foreign_keys=[col],
        back_populates=storage.back_populates,
    )

    if storage.single_parent is not None:
        rel_kwargs['single_parent'] = storage.single_parent

    props[subname] = relationship(arrser, **rel_kwargs)


def _is_array(v):
    return v.Attributes.max_occurs > 1 or issubclass(v, Array)


def _add_array_to_complex(cls, props, subname, subcls, storage):
    arrser_cust = subcls
    if issubclass(subcls, Array):
        arrser_cust, = subcls._type_info.values()

    arrser = arrser_cust
    if arrser_cust.__orig__ is not None:
        arrser = arrser_cust.__orig__

    if storage.multi != False:  # many to many
        _gen_array_m2m(cls, props, subname, arrser, storage)

    elif issubclass(arrser, SimpleModel):  # one to many simple type
        _gen_array_simple(cls, props, subname, arrser_cust, storage)

    else:  # one to many complex type
        _gen_array_o2m(cls, props, subname, arrser, arrser_cust, storage)


def _add_simple_type_to_complex(cls, props, table, subname, subcls, storage,
                                                                    col_kwargs):
    # v has the Attribute values we need whereas real_v is what the
    # user instantiates (thus what sqlalchemy needs)
    if subcls.__orig__ is None:  # vanilla class
        real_v = subcls
    else:  # customized class
        real_v = subcls.__orig__

    assert not getattr(storage, 'multi', False), \
            'Storing a single element-type using a relation table is pointless.'

    assert storage.right is None, \
                               "'right' is ignored in a one-to-one relationship"

    col = _get_col_o2o(cls, subname, subcls, storage.left,
        ondelete=storage.fk_left_ondelete, onupdate=storage.fk_left_onupdate,
        deferrable=storage.fk_left_deferrable,
                                            initially=storage.fk_left_initially)

    storage.left = col.name

    if col.name in table.c:
        col = table.c[col.name]
        if col_kwargs.get('nullable') is False:
            col.nullable = False
    else:
        table.append_column(col)

    rel_kwargs = dict(
        lazy=storage.lazy,
        backref=storage.backref,
        order_by=storage.order_by,
        back_populates=storage.back_populates,
    )

    if storage.single_parent is not None:
        rel_kwargs['single_parent'] = storage.single_parent

    if real_v is (cls.__orig__ or cls):
        (pk_col_name, pk_col_type), = get_pk_columns(cls)
        rel_kwargs['remote_side'] = [table.c[pk_col_name]]

    rel = relationship(real_v, uselist=False, foreign_keys=[col],
        **rel_kwargs)

    _gen_index_info(table, col, subname, subcls)

    props[subname] = rel
    props[col.name] = col


def _add_complex_type_as_table(cls, props, table, subname, subcls, storage,
                                                          col_args, col_kwargs):
    # add one to many relation
    if _is_array(subcls):
        _add_array_to_complex(cls, props, subname, subcls, storage)

    # add one to one relation
    else:
        _add_simple_type_to_complex(cls, props, table, subname, subcls,
                                                            storage, col_kwargs)


def _add_complex_type_as_xml(cls, props, table, subname, subcls, storage,
                                                          col_args, col_kwargs):
    if 'name' in col_kwargs:
        colname = col_kwargs.pop('name')
    else:
        colname = subname

    if colname in table.c:
        col = table.c[colname]
    else:
        t = PGObjectXml(subcls, storage.root_tag, storage.no_ns,
                                                           storage.pretty_print)
        col = Column(colname, t, **col_kwargs)

    props[subname] = col
    if not subname in table.c:
        table.append_column(col)


def _add_complex_type_as_json(cls, props, table, subname, subcls, storage,
                                                     col_args, col_kwargs, dbt):
    if 'name' in col_kwargs:
        colname = col_kwargs.pop('name')
    else:
        colname = subname

    if colname in table.c:
        col = table.c[colname]

    else:
        t = PGObjectJson(subcls, ignore_wrappers=storage.ignore_wrappers,
                                         complex_as=storage.complex_as, dbt=dbt)
        col = Column(colname, t, **col_kwargs)

    props[subname] = col
    if not subname in table.c:
        table.append_column(col)


def _add_complex_type(cls, props, table, subname, subcls):
    if issubclass(subcls, File):
        return _add_file_type(cls, props, table, subname, subcls)

    storage = getattr(subcls.Attributes, 'store_as', None)
    col_args, col_kwargs = sanitize_args(subcls.Attributes.sqla_column_args)
    _sp_attrs_to_sqla_constraints(cls, subcls, col_kwargs)

    if isinstance(storage, c_table):
        return _add_complex_type_as_table(cls, props, table, subname, subcls,
                                                  storage, col_args, col_kwargs)
    if isinstance(storage, c_xml):
        return _add_complex_type_as_xml(cls, props, table, subname, subcls,
                                                  storage, col_args, col_kwargs)
    if isinstance(storage, c_json):
        return _add_complex_type_as_json(cls, props, table, subname, subcls,
                                         storage, col_args, col_kwargs, 'json')
    if isinstance(storage, c_jsonb):
        return _add_complex_type_as_json(cls, props, table, subname, subcls,
                                         storage, col_args, col_kwargs, 'jsonb')
    if isinstance(storage, c_msgpack):
        raise NotImplementedError(c_msgpack)

    if storage is None:
        return

    raise ValueError(storage)


def _convert_fake_table(cls, table):
    metadata = cls.Attributes.sqla_metadata
    table_name = cls.Attributes.table_name

    _table = table
    table_args, table_kwargs = sanitize_args(cls.Attributes.sqla_table_args)
    table = Table(table_name, metadata,
                       *(tuple(table.columns) + table_args), **table_kwargs)

    for index_args, index_kwargs in _table.indexes:
        Index(*index_args, **index_kwargs)

    return table


def _gen_mapper(cls, props, table, cls_bases):
    """Generate SQLAlchemy mapper from Spyne definition data.

    :param cls: La Class.
    :param props: Dict of properties for SQLAlchemt'y Mapper call.
    :param table: A Table instance. Not a `_FakeTable` or anything.
    :param cls_bases: Sequence of class bases.
    """

    inheritance, base_class, base_mapper, inc = _check_inheritance(cls, cls_bases)
    mapper_args, mapper_kwargs = sanitize_args(cls.Attributes.sqla_mapper_args)

    _props = mapper_kwargs.get('properties', None)
    if _props is None:
        mapper_kwargs['properties'] = props
    else:
        props.update(_props)
        mapper_kwargs['properties'] = props

    po = mapper_kwargs.get('polymorphic_on', None)
    if po is not None:
        if not isinstance(po, Column):
            mapper_kwargs['polymorphic_on'] = table.c[po]
        else:
            logger.warning("Deleted invalid 'polymorphic_on' value %r for %r.",
                                                                        po, cls)
            del mapper_kwargs['polymorphic_on']

    if base_mapper is not None:
        mapper_kwargs['inherits'] = base_mapper

    if inheritance is not _SINGLE:
        mapper_args = (table,) + mapper_args

    cls_mapper = mapper(cls, *mapper_args, **mapper_kwargs)

    def on_load(target, context):
        d = target.__dict__

        for k, v in cls.get_flat_type_info(cls).items():
            if not k in d:
                if isclass(v) and issubclass(v, ComplexModelBase):
                    pass
                else:
                    d[k] = None

    event.listen(cls, 'load', on_load)

    return cls_mapper


def _add_file_type(cls, props, table, subname, subcls):
    storage = getattr(subcls.Attributes, 'store_as', None)
    col_args, col_kwargs = sanitize_args(subcls.Attributes.sqla_column_args)
    _sp_attrs_to_sqla_constraints(cls, subcls, col_kwargs)

    if isinstance(storage, HybridFileStore):
        if subname in table.c:
            col = table.c[subname]

        else:
            assert isabs(storage.store)
            #FIXME: Add support for storage markers from spyne.model.complex
            if storage.db_format == 'json':
                t = PGFileJson(storage.store, storage.type)

            elif storage.db_format == 'jsonb':
                t = PGFileJson(storage.store, storage.type, dbt='jsonb')

            else:
                raise NotImplementedError(storage.db_format)

            col = Column(subname, t, **col_kwargs)

        props[subname] = col
        if not subname in table.c:
            table.append_column(col)

    else:
        raise NotImplementedError(storage)


def add_column(cls, subname, subcls):
    """Add field to the given Spyne object also mapped as a SQLAlchemy object
    to a SQLAlchemy table

    :param cls: The class to add the column to.
    :param subname: The column name
    :param subcls: The column type, a ModelBase subclass.
    """

    table = cls.__table__
    mapper_props = {}

    # Add to table
    sqla_type = _get_sqlalchemy_type(subcls)
    if sqla_type is None:  # complex model
        _add_complex_type(cls, mapper_props, table, subname, subcls)
    else:
        _add_simple_type(cls, mapper_props, table, subname, subcls, sqla_type)

    # Add to mapper
    sqla_mapper = cls.Attributes.sqla_mapper
    for subname, subcls in mapper_props.items():
        if not sqla_mapper.has_property(subname):
            sqla_mapper.add_property(subname, subcls)


def _parent_mapper_has_property(cls, cls_bases, k):
    if len(cls_bases) == 0 and cls.__orig__ is cls_bases[0]:
        return False

    for b in cls_bases:
        if not hasattr(b, 'Attributes'):
            continue

        mapper = b.Attributes.sqla_mapper
        if mapper is not None and mapper.has_property(k):
            # print("    Skipping mapping field", "%s.%s" % (cls.__name__, k),
            #          "because parent mapper from", b.__name__, "already has it")
            return True

    # print("NOT skipping mapping field", "%s.%s" % (cls.__name__, k))
    return False


def gen_sqla_info(cls, cls_bases=()):
    """Return SQLAlchemy table object corresponding to the passed Spyne object.
    Also maps given class to the returned table.
    """

    table = _check_table(cls)
    mapper_props = {}

    ancestors = cls.ancestors()
    if len(ancestors) > 0:
        anc_mapper = ancestors[0].Attributes.sqla_mapper
        if anc_mapper is None:
            # no mapper in parent, use all fields
            fields = cls.get_flat_type_info(cls).items()

        elif anc_mapper.concrete:
            # there is mapper in parent and it's concrete, so use all fields
            fields = cls.get_flat_type_info(cls).items()

        else:
            # there is a mapper in parent and it's not concrete, so parent
            # columns are already mapped, so use only own fields.
            fields = cls._type_info.items()

    else:
        # when no parents, use all fields anyway.
        assert set(cls._type_info.items()) == \
               set(cls.get_flat_type_info(cls).items())

        fields = cls.get_flat_type_info(cls).items()

    for k, v in fields:
        if _parent_mapper_has_property(cls, cls_bases, k):
            continue

        t = _get_sqlalchemy_type(v)

        if t is None:  # complex model
            p = getattr(v.Attributes, 'store_as', None)
            if p is None:
                logger.debug("Skipping %s.%s.%s: %r, store_as: %r" % (
                                                cls.get_namespace(),
                                                cls.get_type_name(), k, v, p))
            else:
                _add_complex_type(cls, mapper_props, table, k, v)
        else:
            _add_simple_type(cls, mapper_props, table, k, v, t)

    if isinstance(table, _FakeTable):
        table = _convert_fake_table(cls, table)

    cls_mapper = _gen_mapper(cls, mapper_props, table, cls_bases)

    cls.__tablename__ = cls.Attributes.table_name
    cls.Attributes.sqla_mapper = cls.__mapper__ = cls_mapper
    cls.Attributes.sqla_table = cls.__table__ = table

    return table


def _get_spyne_type(v):
    """Map sqlalchemy types to spyne types."""

    cust = {}
    if v.primary_key:
        cust['primary_key'] = True

    if not v.nullable:
        cust['nullable'] = False
        cust['min_occurs'] = 1

    if isinstance(v.type, sqlalchemy.Enum):
        if v.type.convert_unicode:
            return Unicode(values=v.type.enums, **cust)
        else:
            cust['type_name'] = v.type.name
            return Enum(*v.type.enums, **cust)

    if isinstance(v.type, (sqlalchemy.UnicodeText, sqlalchemy.Text)):
        return Unicode(**cust)

    if isinstance(v.type, (sqlalchemy.Unicode, sqlalchemy.String,
                                                           sqlalchemy.VARCHAR)):
        return Unicode(v.type.length, **cust)

    if isinstance(v.type, sqlalchemy.Numeric):
        return Decimal(v.type.precision, v.type.scale, **cust)

    if isinstance(v.type, PGXml):
        if len(cust) > 0:
            return AnyXml(**cust)
        else:
            return AnyXml

    if isinstance(v.type, PGHtml):
        if len(cust) > 0:
            return AnyHtml(**cust)
        else:
            return AnyHtml

    if type(v.type) in _sq2sp_type_map:
        retval = _sq2sp_type_map[type(v.type)]
        if len(cust) > 0:
            return retval.customize(**cust)
        else:
            return retval

    if isinstance(v.type, (PGObjectJson, PGObjectXml)):
        retval = v.type.cls
        if len(cust) > 0:
            return retval.customize(**cust)
        else:
            return retval

    if isinstance(v.type, PGFileJson):
        retval = v.FileData
        if len(cust) > 0:
            return v.FileData.customize(**cust)
        else:
            return retval

    raise Exception("Spyne type was not found. Probably _sq2sp_type_map "
                                                    "needs a new entry. %r" % v)


def gen_spyne_info(cls):
    table = cls.Attributes.sqla_table
    _type_info = cls._type_info
    mapper_args, mapper_kwargs = sanitize_args(cls.Attributes.sqla_mapper_args)

    if len(_type_info) == 0:
        for c in table.c:
            _type_info[c.name] = _get_spyne_type(c)
    else:
        mapper_kwargs['include_properties'] = _type_info.keys()

    # Map the table to the object
    cls_mapper = mapper(cls, table, *mapper_args, **mapper_kwargs)

    cls.Attributes.table_name = cls.__tablename__ = table.name
    cls.Attributes.sqla_mapper = cls.__mapper__ = cls_mapper


def get_pk_columns(cls):
    """Return primary key fields of a Spyne object."""

    retval = []
    for k, v in cls.get_flat_type_info(cls).items():
        if v.Attributes.sqla_column_args is not None and \
                    v.Attributes.sqla_column_args[-1].get('primary_key', False):
            retval.append((k, v))

    return tuple(retval) if len(retval) > 0 else None




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

import logging
logger = logging.getLogger(__name__)

from collections import deque, defaultdict

import spyne.interface

from spyne import EventManager, MethodDescriptor
from spyne.util import six
from spyne.model import ModelBase, Array, Iterable, ComplexModelBase
from spyne.model.complex import XmlModifier
from spyne.const import xml as namespace


class Interface(object):
    """The ``Interface`` class holds all information needed to build an
    interface document.

    :param app: A :class:`spyne.application.Application` instance.
    """

    def __init__(self, app=None, import_base_namespaces=False):
        self.__ns_counter = 0
        self.__app = None
        self.url = None
        self.classes = {}
        self.imports = {}
        self.service_method_map = {}
        self.method_id_map = {}
        self.nsmap = {}
        self.prefmap = {}
        self.member_methods = deque()
        self.method_descriptor_id_to_key = {}
        self.service_attrs = defaultdict(dict)

        self.import_base_namespaces = import_base_namespaces
        self.app = app

    def set_app(self, value):
        assert self.__app is None, "One interface instance can belong to only " \
                                   "one application instance."

        self.__app = value
        self.reset_interface()
        self.populate_interface()

    def get_app(self):
        return self.__app

    app = property(get_app, set_app)

    @property
    def services(self):
        if self.__app:
            return self.__app.services
        return []

    def reset_interface(self):
        self.classes = {}
        self.imports = {self.get_tns(): set()}
        self.service_method_map = {}
        self.method_id_map = {}
        self.nsmap = dict(namespace.NSMAP)
        self.prefmap = dict(namespace.PREFMAP)
        self.member_methods = deque()

        self.nsmap['tns'] = self.get_tns()
        self.prefmap[self.get_tns()] = 'tns'
        self.deps = defaultdict(set)

    def has_class(self, cls):
        """Returns true if the given class is already included in the interface
        object somewhere."""

        ns = cls.get_namespace()
        tn = cls.get_type_name()

        key = '{%s}%s' % (ns, tn)
        c = self.classes.get(key)
        if c is None:
            return False

        if issubclass(c, ComplexModelBase) and \
                                              issubclass(cls, ComplexModelBase):
            o1 = getattr(cls, '__orig__', None) or cls
            o2 = getattr(c, '__orig__', None) or c

            if o1 is o2:
                return True

            # So that "Array"s and "Iterable"s don't conflict.
            if set((o1, o2)) == set((Array, Iterable)):
                return True

            raise ValueError("classes %r and %r have conflicting names: '%s'" %
                                                                  (cls, c, key))
        return True

    def get_class(self, key):
        """Returns the class definition that corresponds to the given key.
        Keys are in '{namespace}class_name' form.

        Not meant to be overridden.
        """
        return self.classes[key]

    def get_class_instance(self, key):
        """Returns the default class instance that corresponds to the given key.
        Keys are in '{namespace}class_name' form, a.k.a. XML QName format.
        Classes should not enforce arguments to the constructor.

        Not meant to be overridden.
        """
        return self.classes[key]()

    def get_name(self):
        """Returns service name that is seen in the name attribute of the
        definitions tag.

        Not meant to be overridden.
        """

        if self.app:
            return self.app.name

    def get_tns(self):
        """Returns default namespace that is seen in the targetNamespace
        attribute of the definitions tag.

        Not meant to be overridden.
        """
        if self.app:
            return self.app.tns

    def add_method(self, method):
        """Generator method that adds the given method descriptor to the
        interface. Also extracts and yields all the types found in there.

        :param method: A :class:`MethodDescriptor` instance
        :returns: Sequence of :class:`spyne.model.ModelBase` subclasses.
        """

        if not (method.in_header is None):
            if not isinstance(method.in_header, (list, tuple)):
                method.in_header = (method.in_header,)

            for in_header in method.in_header:
                in_header.resolve_namespace(in_header, self.get_tns())
                if method.aux is None:
                    yield in_header
                in_header_ns = in_header.get_namespace()
                if in_header_ns != self.get_tns() and \
                                             self.is_valid_import(in_header_ns):
                    self.imports[self.get_tns()].add(in_header_ns)

        if not (method.out_header is None):
            if not isinstance(method.out_header, (list, tuple)):
                method.out_header = (method.out_header,)

            for out_header in method.out_header:
                out_header.resolve_namespace(out_header, self.get_tns())
                if method.aux is None:
                    yield out_header
                out_header_ns = out_header.get_namespace()
                if out_header_ns != self.get_tns() and \
                                            self.is_valid_import(out_header_ns):
                    self.imports[self.get_tns()].add(out_header_ns)

        if method.faults is None:
            method.faults = []
        elif not (isinstance(method.faults, (list, tuple))):
            method.faults = (method.faults,)

        for fault in method.faults:
            fault.__namespace__ = self.get_tns()
            fault.resolve_namespace(fault, self.get_tns())
            if method.aux is None:
                yield fault

        method.in_message.resolve_namespace(method.in_message, self.get_tns())
        in_message_ns = method.in_message.get_namespace()
        if in_message_ns != self.get_tns() and \
                                            self.is_valid_import(in_message_ns):
            self.imports[self.get_tns()].add(method.in_message.get_namespace())

        if method.aux is None:
            yield method.in_message

        method.out_message.resolve_namespace(method.out_message, self.get_tns())
        assert not method.out_message.get_type_name() is method.out_message.Empty

        out_message_ns = method.out_message.get_namespace()
        if out_message_ns != self.get_tns() and \
                                           self.is_valid_import(out_message_ns):
            self.imports[self.get_tns()].add(out_message_ns)

        if method.aux is None:
            yield method.out_message

        for p in method.patterns:
            p.endpoint = method

    def process_method(self, s, method):
        assert isinstance(method, MethodDescriptor)

        method_key = u'{%s}%s' % (self.app.tns, method.name)

        if issubclass(s, ComplexModelBase):
            method_object_name = method.name.split('.', 1)[0]
            if s.get_type_name() != method_object_name:
                method_key = u'{%s}%s.%s' % (self.app.tns, s.get_type_name(),
                                                                    method.name)

        key = method.gen_interface_key(s)
        if key in self.method_id_map:
            c = self.method_id_map[key].parent_class
            if c is None:
                pass

            elif c is s:
                pass

            elif c.__orig__ is None:
                assert c is s.__orig__, "%r.%s conflicts with %r.%s" % \
                                        (c, key, s.__orig__, key)
            elif s.__orig__ is None:
                assert c.__orig__ is s, "%r.%s conflicts with %r.%s" % \
                                        (c.__orig__, key, s, key)
            else:
                assert c.__orig__ is s.__orig__, "%r.%s conflicts with %r.%s" % \
                                        (c.__orig__, key, s.__orig__, key)
            return

        logger.debug('  adding method %s.%s to match %r tag.',
               method.get_owner_name(s), six.get_function_name(method.function),
                                                                     method_key)

        self.method_id_map[key] = method

        val = self.service_method_map.get(method_key, None)
        if val is None:
            val = self.service_method_map[method_key] = []

        if len(val) == 0:
            val.append(method)

        elif method.aux is not None:
            val.append(method)

        elif val[0].aux is not None:
            val.insert(method, 0)

        else:
            om = val[0]
            os = om.service_class
            if os is None:
                os = om.parent_class
            raise ValueError("\nThe message %r defined in both '%s.%s'"
                                                         " and '%s.%s'"
                                    % (method.name, s.__module__,  s.__name__,
                                                   os.__module__, os.__name__))

    def check_method(self, method):
        """Override this if you need to cherry-pick methods added to the
        interface document."""

        return True

    def populate_interface(self, types=None):
        """Harvests the information stored in individual classes' _type_info
        dictionaries. It starts from function definitions and includes only
        the used objects.
        """

        # populate types
        for s in self.services:
            logger.debug("populating %s types...", s.get_internal_key())

            for method in s.public_methods.values():
                if method.in_header is None:
                    method.in_header = s.__in_header__

                if method.out_header is None:
                    method.out_header = s.__out_header__

                if method.aux is None:
                    method.aux = s.__aux__

                if method.aux is not None:
                    method.aux.methods.append(method.gen_interface_key(s))

                if not self.check_method(method):
                    logger.debug("method %s' discarded by check_method",
                                                               method.class_key)
                    continue

                logger.debug("  enumerating classes for method '%s'",
                                                               method.class_key)
                for cls in self.add_method(method):
                    self.add_class(cls)

        # populate additional types
        for c in self.app.classes:
            self.add_class(c)

        # populate call routes for service methods
        for s in self.services:
            self.service_attrs[s]['tns'] = self.get_tns()
            logger.debug("populating '%s.%s' routes...", s.__module__,
                                                                     s.__name__)
            for method in s.public_methods.values():
                self.process_method(s, method)

        # populate call routes for member methods
        for cls, method in self.member_methods:
            should_we = True
            if method.static_when is not None:
                should_we = method.static_when(self.app)
                logger.debug("static_when returned %r for %s "
                     "while populating methods", should_we, method.internal_key)

            if should_we:
                s = method.service_class
                if s is not None:
                    if method.in_header is None:
                        method.in_header = s.__in_header__

                    if method.out_header is None:
                        method.out_header = s.__out_header__

                    # FIXME: There's no need to process aux info here as it's
                    # not currently known how to write aux member methods in the
                    # first place.

                self.process_method(cls.__orig__ or cls, method)

        # populate method descriptor id to method key map
        self.method_descriptor_id_to_key = dict(((id(v[0]), k)
                                    for k,v in self.service_method_map.items()))

        logger.debug("From this point on, you're not supposed to make any "
                     "changes to the class and method structure of the exposed "
                     "services.")

    tns = property(get_tns)

    def get_namespace_prefix(self, ns):
        """Returns the namespace prefix for the given namespace. Creates a new
        one automatically if it doesn't exist.

        Not meant to be overridden.
        """

        if not (isinstance(ns, str) or isinstance(ns, six.text_type)):
            raise TypeError(ns)

        if not (ns in self.prefmap):
            pref = "s%d" % self.__ns_counter
            while pref in self.nsmap:
                self.__ns_counter += 1
                pref = "s%d" % self.__ns_counter

            self.prefmap[ns] = pref
            self.nsmap[pref] = ns

            self.__ns_counter += 1

        else:
            pref = self.prefmap[ns]

        return pref

    def add_class(self, cls, add_parent=True):
        if self.has_class(cls):
            return

        ns = cls.get_namespace()
        tn = cls.get_type_name()

        assert ns is not None, ('either assign a namespace to the class or call'
                        ' cls.resolve_namespace(cls, "some_default_ns") on it.')

        if not (ns in self.imports) and self.is_valid_import(ns):
            self.imports[ns] = set()

        class_key = '{%s}%s' % (ns, tn)
        logger.debug('    adding class %r for %r', repr(cls), class_key)

        assert class_key not in self.classes, ("Somehow, you're trying to "
            "overwrite %r by %r for class key %r." %
                                      (self.classes[class_key], cls, class_key))

        assert not (cls.get_type_name() is cls.Empty), cls

        self.deps[cls]  # despite the appearances, this is not totally useless.
        self.classes[class_key] = cls
        if ns == self.get_tns():
            self.classes[tn] = cls

        # add parent class
        extends = getattr(cls, '__extends__', None)
        while extends is not None and \
                                   (extends.get_type_name() is ModelBase.Empty):
            extends = getattr(extends, '__extends__', None)

        if add_parent and extends is not None:
            assert issubclass(extends, ModelBase)
            self.deps[cls].add(extends)
            self.add_class(extends)
            parent_ns = extends.get_namespace()
            if parent_ns != ns and not parent_ns in self.imports[ns] and \
                                                self.is_valid_import(parent_ns):
                self.imports[ns].add(parent_ns)
                logger.debug("    importing %r to %r because %r extends %r",
                                            parent_ns, ns, cls.get_type_name(),
                                            extends.get_type_name())

        # add fields
        if issubclass(cls, ComplexModelBase):
            for k, v in cls._type_info.items():
                if v is None:
                    continue

                self.deps[cls].add(v)

                logger.debug("    adding %s.%s = %r", cls.get_type_name(), k, v)
                if v.get_namespace() is None:
                    v.resolve_namespace(v, ns)

                self.add_class(v)

                if v.get_namespace() is None and cls.get_namespace() is not None:
                    v.resolve_namespace(v, cls.get_namespace())

                child_ns = v.get_namespace()
                if child_ns != ns and not child_ns in self.imports[ns] and \
                                                 self.is_valid_import(child_ns):
                    self.imports[ns].add(child_ns)
                    logger.debug("    importing %r to %r for %s.%s(%r)",
                                       child_ns, ns, cls.get_type_name(), k, v)

                if issubclass(v, XmlModifier):
                    self.add_class(v.type)

                    child_ns = v.type.get_namespace()
                    if child_ns != ns and not child_ns in self.imports[ns] and \
                                                 self.is_valid_import(child_ns):
                        self.imports[ns].add(child_ns)
                        logger.debug("    importing %r to %r for %s.%s(%r)",
                                    child_ns, ns, v.get_type_name(), k, v.type)

            if cls.Attributes.methods is not None:
                logger.debug("    populating member methods for '%s.%s'...",
                                       cls.get_namespace(), cls.get_type_name())

                for method_key, descriptor in cls.Attributes.methods.items():
                    assert hasattr(cls, method_key)

                    should_we = True
                    if descriptor.static_when is not None:
                        should_we = descriptor.static_when(self.app)
                        logger.debug("static_when returned %r for %s "
                            "while populating classes",
                                             should_we, descriptor.internal_key)

                    if should_we:
                        self.member_methods.append((cls, descriptor))
                        for c in self.add_method(descriptor):
                            self.add_class(c)

            if cls.Attributes._subclasses is not None:
                logger.debug("    adding subclasses of '%s.%s'...",
                                       cls.get_namespace(), cls.get_type_name())

                for c in cls.Attributes._subclasses:
                    c.resolve_namespace(c, ns)

                    child_ns = c.get_namespace()
                    if child_ns == ns:
                        if not self.has_class(c):
                            self.add_class(c, add_parent=False)
                            self.deps[c].add(cls)
                    else:
                        logger.debug("    not adding %r to %r because it would "
                            "cause circular imports because %r extends %r and "
                            "they don't have the same namespace", child_ns,
                                     ns, c.get_type_name(), cls.get_type_name())

    def is_valid_import(self, ns):
        """This will return False for base namespaces unless told otherwise."""

        if ns is None:
            raise ValueError(ns)

        return self.import_base_namespaces or not (ns in namespace.PREFMAP)


class AllYourInterfaceDocuments(object):
    # AreBelongToUs
    def __init__(self, interface, wsdl11=None):
        self.wsdl11 = wsdl11
        if self.wsdl11 is None and spyne.interface.HAS_WSDL:
            from spyne.interface.wsdl import Wsdl11
            self.wsdl11 = Wsdl11(interface)


class InterfaceDocumentBase(object):
    """Base class for all interface document implementations.

    :param interface: A :class:`spyne.interface.InterfaceBase` instance.
    """

    def __init__(self, interface):
        self.interface = interface
        self.event_manager = EventManager(self)

    def build_interface_document(self):
        """This function is supposed to be called just once, as late as possible
        into the process start. It builds the interface document and caches it
        somewhere. The overriding function should never call the overridden
        function as this may result in the same event firing more than once.
        """

        raise NotImplementedError('Extend and override.')

    def get_interface_document(self):
        """This function is called by server transports that try to satisfy the
        request for the interface document. This should just return a previously
        cached interface document.
        """

        raise NotImplementedError('Extend and override.')




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.server.null`` module contains the NullServer class and its helper
objects.

The name comes from the "null modem connection". Look it up.
"""

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

from spyne import MethodContext, BODY_STYLE_BARE, ComplexModelBase, \
    BODY_STYLE_EMPTY, BODY_STYLE_OUT_BARE, BODY_STYLE_EMPTY_OUT_BARE

from spyne.client import Factory
from spyne.const.ansi_color import LIGHT_RED
from spyne.const.ansi_color import LIGHT_BLUE
from spyne.const.ansi_color import END_COLOR
from spyne.server import ServerBase


_big_header = ('=' * 40) + LIGHT_RED
_big_footer = END_COLOR + ('=' * 40)
_small_header = ('-' * 20) + LIGHT_BLUE
_small_footer = END_COLOR + ('-' * 20)


class NullServer(ServerBase):
    """A server that doesn't support any transport at all -- it's implemented
    to test services without having to run a server.

    Note that:
        1) ``**kwargs`` overwrite ``*args``.
        2) You can do: ::

            logging.getLogger('spyne.server.null').setLevel(logging.CRITICAL)

        to hide context delimiters in logs.
    """

    transport = 'noconn://null.spyne'

    def __init__(self, app, ostr=False, locale='C', appinit=True):
        self.do_appinit = appinit

        super(NullServer, self).__init__(app)

        self.service = _FunctionProxy(self, self.app, is_async=False)
        self.is_async = _FunctionProxy(self, self.app, is_async=True)
        self.factory = Factory(self.app)
        self.ostr = ostr
        self.locale = locale
        self.url = "http://spyne.io/null"

    def appinit(self):
        if self.do_appinit:
            super(NullServer, self).appinit()

    def get_wsdl(self):
        return self.app.get_interface_document(self.url)

    def set_options(self, **kwargs):
        self.service.in_header = kwargs.get('soapheaders',
                                                         self.service.in_header)

    def get_services(self):
        return self.app.interface.service_method_map


class _FunctionProxy(object):
    def __init__(self, server, app, is_async):
        self._app = app
        self._server = server
        self.in_header = None
        self.is_async = is_async

    def __getattr__(self, key):
        return _FunctionCall(self._app, self._server, key, self.in_header,
                          self._server.ostr, self._server.locale, self.is_async)

    def __getitem__(self, key):
        return self.__getattr__(key)


class _FunctionCall(object):
    def __init__(self, app, server, key, in_header, ostr, locale, async_):
        self.app = app

        self._key = key
        self._server = server
        self._in_header = in_header
        self._ostr = ostr
        self._locale = locale
        self._async = async_

    def __call__(self, *args, **kwargs):
        initial_ctx = MethodContext(self, MethodContext.SERVER)
        initial_ctx.method_request_string = self._key
        initial_ctx.in_header = self._in_header
        initial_ctx.transport.type = NullServer.transport
        initial_ctx.locale = self._locale

        contexts = self.app.in_protocol.generate_method_contexts(initial_ctx)

        retval = None
        logger.warning("%s start request %s" % (_big_header, _big_footer))

        if self._async:
            from twisted.internet.defer import Deferred

        for cnt, ctx in enumerate(contexts):
            # this reconstruction is quite costly. I wonder whether it's a
            # problem though.

            _type_info = ctx.descriptor.in_message._type_info
            ctx.in_object = [None] * len(_type_info)
            for i in range(len(args)):
                ctx.in_object[i] = args[i]

            for i, k in enumerate(_type_info.keys()):
                val = kwargs.get(k, None)
                if val is not None:
                    ctx.in_object[i] = val

            if ctx.descriptor.body_style == BODY_STYLE_BARE:
                ctx.in_object = ctx.descriptor.in_message \
                                      .get_serialization_instance(ctx.in_object)

            if cnt == 0:
                p_ctx = ctx
            else:
                ctx.descriptor.aux.initialize_context(ctx, p_ctx, error=None)

            # do logging.getLogger('spyne.server.null').setLevel(logging.CRITICAL)
            # to hide the following
            logger.warning("%s start context %s" % (_small_header,
                                                                 _small_footer))
            logger.info("%r.%r" % (ctx.service_class,
                                                       ctx.descriptor.function))
            try:
                self.app.process_request(ctx)
            finally:
                logger.warning("%s  end context  %s" % (_small_header,
                                                                 _small_footer))

            if cnt == 0:
                if self._async and isinstance(ctx.out_object[0], Deferred):
                    retval = ctx.out_object[0]
                    retval.addCallback(_cb_async, ctx, cnt, self)

                else:
                    retval = _cb_sync(ctx, cnt, self)

        if not self._async:
            p_ctx.close()

        logger.warning("%s  end request  %s" % (_big_header, _big_footer))

        return retval


def _cb_async(ret, ctx, cnt, fc):
    if issubclass(ctx.descriptor.out_message, ComplexModelBase):
        if len(ctx.descriptor.out_message._type_info) == 0:
            ctx.out_object = [None]

        elif len(ctx.descriptor.out_message._type_info) == 1:
            ctx.out_object = [ret]

        else:
            ctx.out_object = ret

    else:
        ctx.out_object = [ret]

    return _cb_sync(ctx, cnt, fc)


def _cb_sync(ctx, cnt, fc):
    retval = None

    if ctx.out_error:
        raise ctx.out_error

    else:
        if ctx.descriptor.is_out_bare():
            retval = ctx.out_object[0]

        elif ctx.descriptor.body_style is BODY_STYLE_EMPTY:
            retval = None

        elif len(ctx.descriptor.out_message._type_info) == 0:
            retval = None

        elif len(ctx.descriptor.out_message._type_info) == 1:
            retval = ctx.out_object[0]

        else:
            retval = ctx.out_object

        if cnt == 0 and fc._ostr:
            fc._server.get_out_string(ctx)
            retval = ctx.out_string

    if cnt > 0:
        ctx.close()

    return retval




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import re
import pytz
import uuid

from math import modf
from time import strptime, mktime
from datetime import timedelta, time, datetime, date
from decimal import Decimal as D, InvalidOperation

from pytz import FixedOffset

try:
    from lxml import etree
    from lxml import html
except ImportError:
    etree = None
    html = None

from spyne.protocol._base import ProtocolMixin
from spyne.model import ModelBase, XmlAttribute, Array, Null, \
    ByteArray, File, ComplexModelBase, AnyXml, AnyHtml, Unicode, String, \
    Decimal, Double, Integer, Time, DateTime, Uuid, Date, Duration, Boolean, Any

from spyne.error import ValidationError

from spyne.model.binary import binary_decoding_handlers, BINARY_ENCODING_USE_DEFAULT

from spyne.util import six
from spyne.model.enum import EnumBase
from spyne.model.primitive.datetime import TIME_PATTERN, DATE_PATTERN

from spyne.util.cdict import cdict


_date_re = re.compile(DATE_PATTERN)
_time_re = re.compile(TIME_PATTERN)
_duration_re = re.compile(
        r'(?P<sign>-?)'
        r'P'
        r'(?:(?P<years>\d+)Y)?'
        r'(?:(?P<months>\d+)M)?'
        r'(?:(?P<days>\d+)D)?'
        r'(?:T(?:(?P<hours>\d+)H)?'
        r'(?:(?P<minutes>\d+)M)?'
        r'(?:(?P<seconds>\d+(.\d+)?)S)?)?'
    )


class InProtocolBase(ProtocolMixin):
    """This is the abstract base class for all input protocol implementations.
    Child classes can implement only the required subset of the public methods.

    An output protocol must implement :func:`serialize` and
    :func:`create_out_string`.

    An input protocol must implement :func:`create_in_document`,
    :func:`decompose_incoming_envelope` and :func:`deserialize`.

    The ProtocolBase class supports the following events:

    * ``before_deserialize``:
      Called before the deserialization operation is attempted.

    * ``after_deserialize``:
      Called after the deserialization operation is finished.

    The arguments the constructor takes are as follows:

    :param app: The application this protocol belongs to.
    :param mime_type: The mime_type this protocol should set for transports
        that support this. This is a quick way to override the mime_type by
        default instead of subclassing the releavant protocol implementation.
    """

    def __init__(self, app=None, validator=None, mime_type=None,
             ignore_wrappers=False, binary_encoding=None, string_encoding=None):

        self.validator = None

        super(InProtocolBase, self).__init__(app=app, mime_type=mime_type,
            ignore_wrappers=ignore_wrappers,
               binary_encoding=binary_encoding, string_encoding=string_encoding)

        self.message = None
        self.validator = None
        self.set_validator(validator)

        if mime_type is not None:
            self.mime_type = mime_type

        fsh = {
            Any: self.any_from_bytes,
            Null: self.null_from_bytes,
            File: self.file_from_bytes,
            Array: self.array_from_bytes,
            Double: self.double_from_bytes,
            String: self.string_from_bytes,
            AnyXml: self.any_xml_from_bytes,
            Boolean: self.boolean_from_bytes,
            Integer: self.integer_from_bytes,
            Unicode: self.unicode_from_bytes,
            AnyHtml: self.any_html_from_bytes,
            ByteArray: self.byte_array_from_bytes,
            EnumBase: self.enum_base_from_bytes,
            ModelBase: self.model_base_from_bytes,
            XmlAttribute: self.xmlattribute_from_bytes,
            ComplexModelBase: self.complex_model_base_from_bytes
        }

        self._from_bytes_handlers = cdict(fsh)
        self._from_unicode_handlers = cdict(fsh)

        self._from_bytes_handlers[Date] = self.date_from_bytes
        self._from_bytes_handlers[Time] = self.time_from_bytes
        self._from_bytes_handlers[Uuid] = self.uuid_from_bytes
        self._from_bytes_handlers[Decimal] = self.decimal_from_bytes
        self._from_bytes_handlers[DateTime] = self.datetime_from_bytes
        self._from_bytes_handlers[Duration] = self.duration_from_bytes

        self._from_unicode_handlers[Date] = self.date_from_unicode
        self._from_unicode_handlers[Uuid] = self.uuid_from_unicode
        self._from_unicode_handlers[Time] = self.time_from_unicode
        self._from_unicode_handlers[Decimal] = self.decimal_from_unicode
        self._from_unicode_handlers[DateTime] = self.datetime_from_unicode
        self._from_unicode_handlers[Duration] = self.duration_from_unicode


        self._datetime_dsmap = {
            None: self._datetime_from_unicode,
            'sec': self._datetime_from_sec,
            'sec_float': self._datetime_from_sec_float,
            'msec': self._datetime_from_msec,
            'msec_float': self._datetime_from_msec_float,
            'usec': self._datetime_from_usec,
        }

    def _datetime_from_sec(self, cls, value):
        try:
            return datetime.fromtimestamp(value)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_sec_float(self, cls, value):
        try:
            return datetime.fromtimestamp(value)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_msec(self, cls, value):
        try:
            return datetime.fromtimestamp(value // 1000)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_msec_float(self, cls, value):
        try:
            return datetime.fromtimestamp(value / 1000)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def _datetime_from_usec(self, cls, value):
        try:
            return datetime.fromtimestamp(value / 1e6)
        except TypeError:
            logger.error("Invalid value %r", value)
            raise

    def create_in_document(self, ctx, in_string_encoding=None):
        """Uses ``ctx.in_string`` to set ``ctx.in_document``."""

    def decompose_incoming_envelope(self, ctx, message):
        """Sets the ``ctx.method_request_string``, ``ctx.in_body_doc``,
        ``ctx.in_header_doc`` and ``ctx.service`` properties of the ctx object,
        if applicable.
        """

    def deserialize(self, ctx, message):
        """Takes a MethodContext instance and a string containing ONE document
        instance in the ``ctx.in_string`` attribute.

        Returns the corresponding native python object in the ctx.in_object
        attribute.
        """

    def validate_document(self, payload):
        """Method to be overriden to perform any sort of custom input
        validation on the parsed input document.
        """

    def set_validator(self, validator):
        """You must override this function if you want your protocol to support
        validation."""

        assert validator is None

        self.validator = None

    def from_bytes(self, class_, string, *args, **kwargs):
        if string is None:
            return None

        if isinstance(string, six.string_types) and \
                           len(string) == 0 and class_.Attributes.empty_is_none:
            return None

        handler = self._from_bytes_handlers[class_]
        return handler(class_, string, *args, **kwargs)

    def from_unicode(self, class_, string, *args, **kwargs):
        if string is None:
            return None

        #if not six.PY2:
        #    assert isinstance(string, str), \
        #        "Invalid type passed to `from_unicode`: {}".format(
        #                                         (class_, type(string), string))

        cls_attrs = self.get_cls_attrs(class_)

        if isinstance(string, six.string_types) and len(string) == 0 and \
                                                        cls_attrs.empty_is_none:
            return None

        handler = self._from_unicode_handlers[class_]
        return handler(class_, string, *args, **kwargs)

    def null_from_bytes(self, cls, value):
        return None

    def any_from_bytes(self, cls, value):
        return value

    def any_xml_from_bytes(self, cls, string):
        try:
            return etree.fromstring(string)
        except etree.XMLSyntaxError as e:
            raise ValidationError(string, "%%r: %r" % e)

    def any_html_from_bytes(self, cls, string):
        try:
            return html.fromstring(string)
        except etree.ParserError as e:
            if e.args[0] == "Document is empty":
                pass
            else:
                raise

    def uuid_from_unicode(self, cls, string, suggested_encoding=None):
        attr = self.get_cls_attrs(cls)
        ser_as = attr.serialize_as
        encoding = attr.encoding

        if encoding is None:
            encoding = suggested_encoding

        retval = string

        if ser_as in ('bytes', 'bytes_le'):
            retval, = binary_decoding_handlers[encoding](string)

        try:
            retval = _uuid_deserialize[ser_as](retval)
        except ValueError as e:
            raise ValidationError(e)

        return retval

    def uuid_from_bytes(self, cls, string, suggested_encoding=None, **_):
        attr = self.get_cls_attrs(cls)
        ser_as = attr.serialize_as
        encoding = attr.encoding

        if encoding is None:
            encoding = suggested_encoding

        retval = string

        if ser_as in ('bytes', 'bytes_le'):
            retval, = binary_decoding_handlers[encoding](string)
        elif isinstance(string, six.binary_type):
            retval = string.decode('ascii')

        try:
            retval = _uuid_deserialize[ser_as](retval)
        except ValueError as e:
            raise ValidationError(e)

        return retval

    def unicode_from_bytes(self, cls, value):
        retval = value

        if isinstance(value, six.binary_type):
            cls_attrs = self.get_cls_attrs(cls)

            if cls_attrs.encoding is not None:
                retval = six.text_type(value, cls_attrs.encoding,
                                                errors=cls_attrs.unicode_errors)

            elif self.string_encoding is not None:
                retval = six.text_type(value, self.string_encoding,
                                                errors=cls_attrs.unicode_errors)

            else:
                retval = six.text_type(value, errors=cls_attrs.unicode_errors)

        return retval

    def string_from_bytes(self, cls, value):
        retval = value
        cls_attrs = self.get_cls_attrs(cls)
        if isinstance(value, six.text_type):
            if cls_attrs.encoding is None:
                raise Exception("You need to define a source encoding for "
                                "decoding incoming unicode values.")
            else:
                retval = value.encode(cls_attrs.encoding)

        return retval

    def decimal_from_unicode(self, cls, string):
        cls_attrs = self.get_cls_attrs(cls)
        if cls_attrs.max_str_len is not None and len(string) > \
                                                     cls_attrs.max_str_len:
            raise ValidationError(string, "Decimal %%r longer than %d "
                                          "characters" % cls_attrs.max_str_len)

        try:
            return D(string)
        except InvalidOperation as e:
            raise ValidationError(string, "%%r: %r" % e)

    def decimal_from_bytes(self, cls, string):
        return self.decimal_from_unicode(cls,
                                    string.decode(self.default_string_encoding))

    def double_from_bytes(self, cls, string):
        try:
            return float(string)
        except (TypeError, ValueError) as e:
            raise ValidationError(string, "%%r: %r" % e)

    def integer_from_bytes(self, cls, string):
        cls_attrs = self.get_cls_attrs(cls)

        if isinstance(string, (six.text_type, six.binary_type)) and \
                                    cls_attrs.max_str_len is not None and \
                                    len(string) > cls_attrs.max_str_len:
            raise ValidationError(string,
                                         "Integer %%r longer than %d characters"
                                                        % cls_attrs.max_str_len)

        try:
            return int(string)
        except ValueError:
            raise ValidationError(string, "Could not cast %r to integer")

    def time_from_unicode(self, cls, string):
        """Expects ISO formatted times."""

        match = _time_re.match(string)
        if match is None:
            raise ValidationError(string, "%%r does not match regex %r " %
                                                               _time_re.pattern)

        fields = match.groupdict(0)
        microsec = fields.get('sec_frac')
        if microsec is None or microsec == 0:
            microsec = 0
        else:
            microsec = min(999999, int(round(float(microsec) * 1e6)))

        return time(int(fields['hr']), int(fields['min']),
                                                   int(fields['sec']), microsec)

    def time_from_bytes(self, cls, string):
        if isinstance(string, six.binary_type):
            string = string.decode(self.default_string_encoding)

        return self.time_from_unicode(cls, string)

    def date_from_unicode_iso(self, cls, string):
        """This is used by protocols like SOAP who need ISO8601-formatted dates
        no matter what.
        """

        try:
            return date(*(strptime(string, u'%Y-%m-%d')[0:3]))

        except ValueError:
            match = cls._offset_re.match(string)

            if match:
                year = int(match.group('year'))
                month = int(match.group('month'))
                day = int(match.group('day'))

                return date(year, month, day)

            raise ValidationError(string)

    def enum_base_from_bytes(self, cls, value):
        if self.validator is self.SOFT_VALIDATION and not (
                                        cls.validate_string(cls, value)):
            raise ValidationError(value)
        return getattr(cls, value)

    def model_base_from_bytes(self, cls, value):
        return cls.from_bytes(value)

    def datetime_from_unicode_iso(self, cls, string):
        astz = self.get_cls_attrs(cls).as_timezone

        match = cls._utc_re.match(string)
        if match:
            tz = pytz.utc
            retval = _parse_datetime_iso_match(match, tz=tz)
            if astz is not None:
                retval = retval.astimezone(astz)
            return retval

        if match is None:
            match = cls._offset_re.match(string)
            if match:
                tz_hr, tz_min = [int(match.group(x))
                                                   for x in ("tz_hr", "tz_min")]
                tz = FixedOffset(tz_hr * 60 + tz_min, {})
                retval = _parse_datetime_iso_match(match, tz=tz)
                if astz is not None:
                    retval = retval.astimezone(astz)
                return retval

        if match is None:
            match = cls._local_re.match(string)
            if match:
                retval = _parse_datetime_iso_match(match)
                if astz:
                    retval = retval.replace(tzinfo=astz)
                return retval

        raise ValidationError(string)

    def datetime_from_unicode(self, cls, string):
        serialize_as = self.get_cls_attrs(cls).serialize_as
        return self._datetime_dsmap[serialize_as](cls, string)

    def datetime_from_bytes(self, cls, string):
        if isinstance(string, six.binary_type):
            string = string.decode(self.default_string_encoding)

        serialize_as = self.get_cls_attrs(cls).serialize_as
        return self._datetime_dsmap[serialize_as](cls, string)

    def date_from_bytes(self, cls, string):
        if isinstance(string, six.binary_type):
            string = string.decode(self.default_string_encoding)

        date_format = self._get_date_format(self.get_cls_attrs(cls))
        try:
            if date_format is not None:
                dt = datetime.strptime(string, date_format)
                return date(dt.year, dt.month, dt.day)

            return self.date_from_unicode_iso(cls, string)

        except ValueError as e:
            match = cls._offset_re.match(string)
            if match:
                return date(int(match.group('year')),
                            int(match.group('month')), int(match.group('day')))
            else:
                raise ValidationError(string,
                                         "%%r: %s" % repr(e).replace("%", "%%"))

    def date_from_unicode(self, cls, string):
        date_format = self._get_date_format(self.get_cls_attrs(cls))
        try:
            if date_format is not None:
                dt = datetime.strptime(string, date_format)
                return date(dt.year, dt.month, dt.day)

            return self.date_from_unicode_iso(cls, string)

        except ValueError as e:
            match = cls._offset_re.match(string)
            if match:
                return date(int(match.group('year')),
                            int(match.group('month')), int(match.group('day')))
            else:
                # the message from ValueError is quite nice already
                raise ValidationError(e.message, "%s")

    def duration_from_unicode(self, cls, string):
        duration = _duration_re.match(string).groupdict(0)
        if duration is None:
            raise ValidationError(string,
                "Time data '%%s' does not match regex '%s'" %
                                                        (_duration_re.pattern,))

        days = int(duration['days'])
        days += int(duration['months']) * 30
        days += int(duration['years']) * 365
        hours = int(duration['hours'])
        minutes = int(duration['minutes'])
        seconds = float(duration['seconds'])
        f, i = modf(seconds)
        seconds = i
        microseconds = int(1e6 * f)

        delta = timedelta(days=days, hours=hours, minutes=minutes,
            seconds=seconds, microseconds=microseconds)

        if duration['sign'] == "-":
            delta *= -1

        return delta

    def duration_from_bytes(self, cls, string):
        if isinstance(string, six.binary_type):
            string = string.decode(self.default_string_encoding)

        return self.duration_from_unicode(cls, string)

    def boolean_from_bytes(self, cls, string):
        return string.lower() in ('true', '1')

    def byte_array_from_bytes(self, cls, value, suggested_encoding=None):
        encoding = self.get_cls_attrs(cls).encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            encoding = suggested_encoding
        return binary_decoding_handlers[encoding](value)

    def file_from_bytes(self, cls, value, suggested_encoding=None):
        encoding = self.get_cls_attrs(cls).encoding
        if encoding is BINARY_ENCODING_USE_DEFAULT:
            encoding = suggested_encoding

        return File.Value(data=binary_decoding_handlers[encoding](value))

    def complex_model_base_from_bytes(self, cls, string, **_):
        raise TypeError("Only primitives can be deserialized from string.")

    def array_from_bytes(self, cls, string, **_):
        if self.get_cls_attrs(cls).serialize_as != 'sd-list':
            raise TypeError("Only primitives can be deserialized from string.")

        # sd-list being space-delimited list.
        retval = []
        inner_type, = cls._type_info.values()
        for s in string.split():
            retval.append(self.from_bytes(inner_type, s))

        return retval

    def xmlattribute_from_bytes(self, cls, value):
        return self.from_bytes(cls.type, value)

    def _datetime_from_unicode(self, cls, string):
        cls_attrs = self.get_cls_attrs(cls)

        # get parser
        parser = cls_attrs.parser

        # get date_format
        dt_format = cls_attrs.dt_format
        if dt_format is None:
            dt_format = cls_attrs.date_format
        if dt_format is None:
            dt_format = cls_attrs.out_format
        if dt_format is None:
            dt_format = cls_attrs.format

        # parse the string
        if parser is not None:
            retval = parser(self, cls, string)

        elif dt_format is not None:
            if six.PY2:
                # FIXME: perhaps it should encode to string's encoding instead
                # of utf8 all the time
                if isinstance(dt_format, six.text_type):
                    dt_format = dt_format.encode('utf8')
                if isinstance(string, six.text_type):
                    string = string.encode('utf8')

            retval = datetime.strptime(string, dt_format)

            astz = cls_attrs.as_timezone
            if astz:
                retval = retval.astimezone(cls_attrs.as_time_zone)

        else:
            retval = self.datetime_from_unicode_iso(cls, string)

        return retval


_uuid_deserialize = {
    None: lambda s: uuid.UUID(s),
    'hex': lambda s: uuid.UUID(hex=s),
    'urn': lambda s: uuid.UUID(hex=s),
    'bytes': lambda s: uuid.UUID(bytes=s),
    'bytes_le': lambda s: uuid.UUID(bytes_le=s),
    'fields': lambda s: uuid.UUID(fields=s),
    'int': lambda s: uuid.UUID(int=s),
    ('int', int): lambda s: uuid.UUID(int=s),
    ('int', str): lambda s: uuid.UUID(int=int(s)),
}

if six.PY2:
    _uuid_deserialize[('int', long)] = _uuid_deserialize[('int', int)]


def _parse_datetime_iso_match(date_match, tz=None):
    fields = date_match.groupdict()

    year = int(fields.get('year'))
    month = int(fields.get('month'))
    day = int(fields.get('day'))
    hour = int(fields.get('hr'))
    minute = int(fields.get('min'))
    second = int(fields.get('sec'))
    usecond = fields.get("sec_frac")
    if usecond is None:
        usecond = 0
    else:
        # we only get the most significant 6 digits because that's what
        # datetime can handle.
        usecond = min(999999, int(round(float(usecond) * 1e6)))

    return datetime(year, month, day, hour, minute, second, usecond, tz)


_dt_sec = lambda cls, val: \
        int(mktime(val.timetuple()))
_dt_sec_float = lambda cls, val: \
        mktime(val.timetuple()) + (val.microsecond / 1e6)

_dt_msec = lambda cls, val: \
        int(mktime(val.timetuple())) * 1000 + (val.microsecond // 1000)
_dt_msec_float = lambda cls, val: \
        mktime(val.timetuple()) * 1000 + (val.microsecond / 1000.0)

_dt_usec = lambda cls, val: \
        int(mktime(val.timetuple())) * 1000000 + val.microsecond

_datetime_smap = {
    'sec': _dt_sec,
    'secs': _dt_sec,
    'second': _dt_sec,
    'seconds': _dt_sec,

    'sec_float': _dt_sec_float,
    'secs_float': _dt_sec_float,
    'second_float': _dt_sec_float,
    'seconds_float': _dt_sec_float,

    'msec': _dt_msec,
    'msecs': _dt_msec,
    'msecond': _dt_msec,
    'mseconds': _dt_msec,
    'millisecond': _dt_msec,
    'milliseconds': _dt_msec,

    'msec_float': _dt_msec_float,
    'msecs_float': _dt_msec_float,
    'msecond_float': _dt_msec_float,
    'mseconds_float': _dt_msec_float,
    'millisecond_float': _dt_msec_float,
    'milliseconds_float': _dt_msec_float,

    'usec': _dt_usec,
    'usecs': _dt_usec,
    'usecond': _dt_usec,
    'useconds': _dt_usec,
    'microsecond': _dt_usec,
    'microseconds': _dt_usec,
}


def _file_to_iter(f):
    try:
        data = f.read(65536)
        while len(data) > 0:
            yield data
            data = f.read(65536)

    finally:
        f.close()



# encoding: utf-8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""Useful stuff to integrate Spyne with Django.

* Django model <-> spyne type mapping
* Service for common exception handling

"""

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import re

from itertools import chain

from django.core.exceptions import (ImproperlyConfigured, ObjectDoesNotExist,
                                    ValidationError as DjValidationError)
from django.core.validators import (slug_re,
                                    MinLengthValidator, MaxLengthValidator)
try:
    from django.core.validators import comma_separated_int_list_re
except ImportError:
    comma_separated_int_list_re = re.compile(r'^[\d,]+$')

from spyne.error import (ResourceNotFoundError, ValidationError as
                         BaseValidationError, Fault)
from spyne.model import primitive
from spyne.model.complex import ComplexModelMeta, ComplexModelBase
from spyne.service import Service
from spyne.util.cdict import cdict
from spyne.util.odict import odict
from spyne.util.six import add_metaclass


# regex is based on http://www.w3.org/TR/xforms20/#xforms:email
email_re = re.compile(
    r"[A-Za-z0-9!#-'\*\+\-/=\?\^_`\{-~]+"
    r"(\.[A-Za-z0-9!#-'\*\+\-/=\?\^_`\{-~]+)*@"
    # domain part is either a single symbol
    r"(([a-zA-Z0-9]|"
    # or have at least two symbols
    # hyphen can't be at the beginning or end of domain part
    # domain should contain at least 2 parts, the last one is TLD
    r"([a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])+)\.)+"
    # TLD should contain only letters, at least 2
    r"[A-Za-z]{2,}", re.IGNORECASE)


def _handle_minlength(validator, params):
    new_min = validator.limit_value
    old_min = params.setdefault('min_len', new_min)
    params['min_len'] = max(old_min, new_min)


def _handle_maxlength(validator, params):
    new_max = validator.limit_value
    old_max = params.setdefault('max_len', new_max)
    params['max_len'] = min(old_max, new_max)


class BaseDjangoFieldMapper(object):

    """Abstrace base class for field mappers."""

    _VALIDATOR_HANDLERS = cdict({
        MinLengthValidator: _handle_minlength,
        MaxLengthValidator: _handle_maxlength,
    })

    @staticmethod
    def is_field_nullable(field, **kwargs):
        """Return True if django field is nullable."""
        return field.null

    @staticmethod
    def is_field_blank(field, **kwargs):
        """Return True if django field is blank."""
        return field.blank

    def map(self, field, **kwargs):
        """Map field to spyne model.

        :param field: Django Field instance
        :param kwargs: Extra params to configure spyne model
        :returns: tuple (field attribute name, mapped spyne model)

        """
        params = kwargs.copy()

        self._process_validators(field.validators, params)

        nullable = self.is_field_nullable(field, **kwargs)
        blank = self.is_field_blank(field, **kwargs)
        required = not (field.has_default() or blank or field.primary_key)

        if field.has_default():
            params['default'] = field.get_default()

        spyne_model = self.get_spyne_model(field, **kwargs)
        customized_model = spyne_model(nullable=nullable,
                                       min_occurs=int(required), **params)

        return (field.attname, customized_model)

    def get_spyne_model(self, field, **kwargs):
        """Return spyne model for given Django field."""
        raise NotImplementedError

    def _process_validators(self, validators, params):
        for v in validators:
            handler = self._VALIDATOR_HANDLERS.get(type(v))
            if handler:
                handler(v, params)


class DjangoFieldMapper(BaseDjangoFieldMapper):

    """Basic mapper for django fields."""

    def __init__(self, spyne_model):
        """Django field mapper constructor."""
        self.spyne_model = spyne_model

    def get_spyne_model(self, field, **kwargs):
        """Return configured spyne model."""
        return self.spyne_model


class DecimalMapper(DjangoFieldMapper):

    """Mapper for DecimalField."""

    def map(self, field, **kwargs):
        """Map DecimalField to spyne model.

        :returns: tuple (field attribute name, mapped spyne model)

        """
        params = kwargs.copy()
        params.update({
            'total_digits': field.max_digits,
            'fraction_digits': field.decimal_places,
        })
        return super(DecimalMapper, self).map(field, **params)


class RelationMapper(BaseDjangoFieldMapper):

    """Mapper for relation fields (ForeignKey, OneToOneField)."""

    def __init__(self, django_model_mapper):
        """Constructor for relation field mapper."""
        self.django_model_mapper = django_model_mapper

    @staticmethod
    def is_field_blank(field, **kwargs):
        """Return True if `optional_relations` is set.

        Otherwise use basic behaviour.

        """
        optional_relations = kwargs.get('optional_relations', False)
        return (optional_relations or
                BaseDjangoFieldMapper.is_field_blank(field, **kwargs))

    def get_spyne_model(self, field, **kwargs):
        """Return spyne model configured by related field."""
        related_field = field.rel.get_related_field() if hasattr(field, 'rel') else field.remote_field.get_related_field()
        field_type = related_field.__class__.__name__
        field_mapper = self.django_model_mapper.get_field_mapper(field_type)

        _, related_spyne_model = field_mapper.map(related_field, **kwargs)
        return related_spyne_model


class DjangoModelMapper(object):

    r"""Mapper from django models to spyne complex models.

    You can extend it registering new field types: ::

        class NullBooleanMapper(DjangoFieldMapper):

            def map(self, field, **kwargs):
                params = kwargs.copy()
                # your mapping logic goes here
                return super(NullBooleanMapper, self).map(field, **params)

        default_model_mapper.register_field_mapper('NullBooleanField', \
                NullBooleanMapper(primitive.Boolean))


    You may subclass it if you want different mapping logic for different
    Django models.

    """

    field_mapper_class = DjangoFieldMapper

    class UnknownFieldMapperException(Exception):

        """Raises when there is no field mapper for given django_type."""

    def __init__(self, django_spyne_models=()):
        """Register field mappers in internal registry."""
        self._registry = {}

        for django_type, spyne_model in django_spyne_models:
            self.register(django_type, spyne_model)

    def get_field_mapper(self, django_type):
        """Get mapper registered for given django_type.

        :param django_type: Django internal field type
        :returns: registered mapper
        :raises: :exc:`UnknownFieldMapperException`

        """
        try:
            return self._registry[django_type]
        except KeyError:
            raise self.UnknownFieldMapperException(
                'No mapper for field type {0}'.format(django_type))

    def register(self, django_type, spyne_model):
        """Register default field mapper for django_type and spyne_model.

        :param django_type: Django internal field type
        :param spyne_model: Spyne model, usually primitive

        """
        field_mapper = self.field_mapper_class(spyne_model)
        self.register_field_mapper(django_type, field_mapper)

    def register_field_mapper(self, django_type, field_mapper):
        """Register field mapper for django_type.

        :param django_type: Django internal field type
        :param field_mapper: :class:`DjangoFieldMapper` instance

        """
        self._registry[django_type] = field_mapper

    @staticmethod
    def get_all_field_names(meta):
        if hasattr(meta, 'get_all_field_names'):
            return meta.get_all_field_names()

        return list(set(chain.from_iterable(
            (field.name, field.attname) if hasattr(field, 'attname') else (
            field.name,)
            for field in meta.get_fields()
            # For complete backwards compatibility, you may want to exclude
            # GenericForeignKey from the results.
            if not (field.many_to_one and field.related_model is None)
        )))

    @staticmethod
    def _get_fields(django_model, exclude=None):
        field_names = set(exclude) if exclude is not None else set()
        meta = django_model._meta  # pylint: disable=W0212
        unknown_fields_names = \
             field_names.difference(DjangoModelMapper.get_all_field_names(meta))

        if unknown_fields_names:
            raise ImproperlyConfigured(
                'Unknown field names: {0}'
                .format(', '.join(unknown_fields_names)))

        return [field for field in meta.fields if field.name not in
                field_names]

    def map(self, django_model, exclude=None, **kwargs):
        """Prepare dict of model fields mapped to spyne models.

        :param django_model: Django model class.
        :param exclude: list of fields excluded from mapping.
        :param kwargs: extra kwargs are passed to all field mappers

        :returns: dict mapping attribute names to spyne models
        :raises: :exc:`UnknownFieldMapperException`

        """
        field_map = odict()

        for field in self._get_fields(django_model, exclude):
            field_type = field.__class__.__name__

            try:
                field_mapper = self._registry[field_type]
            except KeyError:
                # mapper for this field is not registered
                if not (field.has_default() or field.null):
                    # field is required
                    raise self.UnknownFieldMapperException(
                        'No mapper for field type {0}'.format(field_type))
                else:
                    # skip this field
                    logger.info('Field {0} is skipped from mapping.')
                    continue

            attr_name, spyne_model = field_mapper.map(field, **kwargs)
            field_map[attr_name] = spyne_model

        return field_map


def strip_regex_metachars(pattern):
    """Strip ^ and $ from pattern begining and end.

    According to http://www.w3.org/TR/xmlschema-0/#regexAppendix XMLSchema
    expression language does not contain the metacharacters ^ and $.

    :returns: stripped pattern string

    """
    start = 0
    till = len(pattern)

    if pattern.startswith('^'):
        start = 1

    if pattern.endswith('$'):
        till -= 1

    return pattern[start:till]


# django's own slug_re.pattern is invalid according to xml schema -- it doesn't
# like the location of the dash character. using the equivalent pattern accepted
# by xml schema here.
SLUG_RE_PATTERN = '[a-zA-Z0-9_-]+'


DEFAULT_FIELD_MAP = (
    ('AutoField', primitive.Integer32),
    ('CharField', primitive.NormalizedString),
    ('SlugField', primitive.Unicode(
        type_name='Slug', pattern=strip_regex_metachars(SLUG_RE_PATTERN))),
    ('TextField', primitive.Unicode),
    ('EmailField', primitive.Unicode(
        type_name='Email', pattern=strip_regex_metachars(email_re.pattern))),
    ('CommaSeparatedIntegerField', primitive.Unicode(
        type_name='CommaSeparatedField',
        pattern=strip_regex_metachars(comma_separated_int_list_re.pattern))),
    ('URLField', primitive.AnyUri),
    ('FilePathField', primitive.Unicode),

    ('BooleanField', primitive.Boolean),
    ('NullBooleanField', primitive.Boolean),
    ('IntegerField', primitive.Integer),
    ('BigIntegerField', primitive.Integer64),
    ('PositiveIntegerField', primitive.UnsignedInteger32),
    ('SmallIntegerField', primitive.Integer16),
    ('PositiveSmallIntegerField', primitive.UnsignedInteger16),
    ('FloatField', primitive.Double),

    ('TimeField', primitive.Time),
    ('DateField', primitive.Date),
    ('DateTimeField', primitive.DateTime),

    # simple fixed defaults for relation fields
    ('ForeignKey', primitive.Integer32),
    ('OneToOneField', primitive.Integer32),
)


def model_mapper_factory(mapper_class, field_map):
    """Factory for model mappers.

    The factory is useful to create custom field mappers based on default one.

    """
    model_mapper = mapper_class(field_map)

    # register relation field mappers that are aware of related field type
    model_mapper.register_field_mapper(
        'ForeignKey', RelationMapper(model_mapper))

    model_mapper.register_field_mapper(
        'OneToOneField', RelationMapper(model_mapper))

    model_mapper.register_field_mapper('DecimalField',
                                       DecimalMapper(primitive.Decimal))
    return model_mapper


default_model_mapper = model_mapper_factory(DjangoModelMapper,
                                            DEFAULT_FIELD_MAP)


class DjangoComplexModelMeta(ComplexModelMeta):

    """Meta class for complex spyne models representing Django models."""

    def __new__(mcs, name, bases, attrs):  # pylint: disable=C0202
        """Populate new complex type from configured Django model."""
        super_new = super(DjangoComplexModelMeta, mcs).__new__

        abstract = bool(attrs.get('__abstract__', False))

        if abstract:
            # skip processing of abstract models
            return super_new(mcs, name, bases, attrs)

        attributes = attrs.get('Attributes')

        if attributes is None:
            raise ImproperlyConfigured('You have to define Attributes and '
                                       'specify Attributes.django_model')

        if getattr(attributes, 'django_model', None) is None:
            raise ImproperlyConfigured('You have to define django_model '
                                       'attribute in Attributes')

        mapper = getattr(attributes, 'django_mapper', default_model_mapper)
        attributes.django_mapper = mapper
        exclude = getattr(attributes, 'django_exclude', None)
        optional_relations = getattr(attributes, 'django_optional_relations',
                                     False)
        spyne_attrs = mapper.map(attributes.django_model, exclude=exclude,
                                 optional_relations=optional_relations)
        spyne_attrs.update(attrs)
        return super_new(mcs, name, bases, spyne_attrs)


@add_metaclass(DjangoComplexModelMeta)
class DjangoComplexModel(ComplexModelBase):

    """Base class with Django model mapping support.

    Sample usage: ::

        class PersonType(DjangoComplexModel):
            class Attributes(DjangoComplexModel.Attributes):
                django_model = Person


    Attribute :attr:`django_model` is required for Django model mapping
    machinery. You can customize your types defining custom type fields: ::

        class PersonType(DjangoComplexModel):
            gender = primitive.Unicode(pattern='^[FM]$')

            class Attributes(DjangoComplexModel.Attributes):
                django_model = Person


    There is an option to specify custom mapper: ::

        class PersonType(DjangoComplexModel):
            class Attributes(DjangoComplexModel.Attributes):
                django_model = Person
                django_mapper = my_custom_mapper

    You can also exclude some fields from mapping: ::

        class PersonType(DjangoComplexModel):
            class Attributes(DjangoComplexModel.Attributes):
                django_model = Person
                django_exclude = ['phone']

    You may set `django_optional_relations`` attribute flag to indicate
    that relation fields (ForeignKey, OneToOneField) of your model are
    optional.  This is useful when you want to create base and related
    instances in remote procedure. In this case primary key of base model is
    not yet available.

    """

    __abstract__ = True


class ObjectNotFoundError(ResourceNotFoundError):

    """Fault constructed from `model.DoesNotExist` exception."""

    def __init__(self, does_not_exist_exc):
        """Construct fault with code Client.<object_name>NotFound."""
        message = str(does_not_exist_exc)
        object_name = message.split()[0]
        # we do not want to reuse initialization of ResourceNotFoundError
        Fault.__init__(
            self, faultcode='Client.{0}NotFound'.format(object_name),
            faultstring=message)


class ValidationError(BaseValidationError):

    """Fault constructed from `ValidationError` exception."""

    def __init__(self, validation_error_exc):
        """Construct fault with code Client.<validation_error_type_name>."""
        message = str(validation_error_exc)
        # we do not want to reuse initialization of BaseValidationError
        Fault.__init__(
            self, faultcode='Client.{0}'.format(
                type(validation_error_exc).__name__), faultstring=message)


class DjangoService(Service):

    """Service with common Django exception handling."""

    @classmethod
    def call_wrapper(cls, ctx):
        """Handle common Django exceptions."""
        try:
            out_object = super(DjangoService, cls).call_wrapper(ctx)
        except ObjectDoesNotExist as e:
            raise ObjectNotFoundError(e)
        except DjValidationError as e:
            raise ValidationError(e)
        return out_object


# FIXME: To be removed in Spyne 3
DjangoServiceBase = DjangoService




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import absolute_import


try:
    import colorama
    R = lambda s: ''.join((colorama.Fore.RED, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))
    G = lambda s: ''.join((colorama.Fore.GREEN, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))
    B = lambda s: ''.join((colorama.Fore.BLUE, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))

    DARK_R = lambda s: ''.join((colorama.Fore.RED, s, colorama.Style.RESET_ALL))
    DARK_G = lambda s: ''.join((colorama.Fore.GREEN, s, colorama.Style.RESET_ALL))
    DARK_B = lambda s: ''.join((colorama.Fore.BLUE, s, colorama.Style.RESET_ALL))

    YEL = lambda s: ''.join((colorama.Fore.YELLOW, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))
    MAG = lambda s: ''.join((colorama.Fore.MAGENTA, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))
    CYA = lambda s: ''.join((colorama.Fore.CYAN, colorama.Style.BRIGHT, s,
                                                    colorama.Style.RESET_ALL))

    DARK_YEL = lambda s: ''.join((colorama.Fore.YELLOW, s,
                                                      colorama.Style.RESET_ALL))
    DARK_MAG = lambda s: ''.join((colorama.Fore.MAGENTA, s,
                                                      colorama.Style.RESET_ALL))
    DARK_CYA = lambda s: ''.join((colorama.Fore.CYAN, s,
                                                      colorama.Style.RESET_ALL))

except ImportError:
    R = lambda s: s
    G = lambda s: s
    B = lambda s: s
    DARK_R = lambda s: s
    DARK_G = lambda s: s
    DARK_B = lambda s: s
    YEL = lambda s: s
    MAG = lambda s: s
    CYA = lambda s: s
    DARK_YEL = lambda s: s
    DARK_MAG = lambda s: s
    DARK_CYA = lambda s: s


if __name__ == '__main__':
    print(R("RED"))
    print(G("GREEN"))
    print(B("BLUE"))
    print(DARK_R("DARK_RED"))
    print(DARK_G("DARK_GREEN"))
    print(DARK_B("DARK_BLUE"))
    print(YEL("YELLOW"))
    print(MAG("MAGENTA"))
    print(CYA("CYAN"))



# encoding: utf-8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""Metaclass utilities for
:attr:`spyne.model.complex.ComplexModelBase.Attributes.declare_order`
"""

import sys
import inspect

from functools import wraps
from itertools import chain
from warnings import warn

from spyne.util.odict import odict


class ClassNotFoundException(Exception):
    """Raise when class declaration is not found in frame stack."""


class AttributeNotFoundException(Exception):
    """Raise when attribute is not found in class declaration."""


class Prepareable(type):
    """Implement __prepare__ for Python 2.

    This class is used in Python 2 and Python 3 to support `six.add_metaclass`
    decorator that populates attributes of resulting class from plain unordered
    attributes dict of decorated class.

    Based on https://gist.github.com/DasIch/5562625
    """

    def __new__(cls, name, bases, attributes):
        try:
            constructor = attributes["__new__"]
        except KeyError:
            return type.__new__(cls, name, bases, attributes)

        def preparing_constructor(cls, name, bases, attributes):
            # Don't bother with this shit unless the user *explicitly* asked for
            # it
            for c in chain(bases, [cls]):
                if hasattr(c,'Attributes') and not \
                               (c.Attributes.declare_order in (None, 'random')):
                    break
            else:
                return constructor(cls, name, bases, attributes)

            try:
                cls.__prepare__
            except AttributeError:
                return constructor(cls, name, bases, attributes)

            if isinstance(attributes, odict):
                # we create class dynamically with passed odict
                return constructor(cls, name, bases, attributes)

            current_frame = sys._getframe()
            class_declaration = None

            while class_declaration is None:
                literals = list(reversed(current_frame.f_code.co_consts))

                for literal in literals:
                    if inspect.iscode(literal) and literal.co_name == name:
                        class_declaration = literal
                        break

                else:
                    if current_frame.f_back:
                        current_frame = current_frame.f_back
                    else:
                        raise ClassNotFoundException(
                            "Can't find class declaration in any frame")

            def get_index(attribute_name,
                            _names=class_declaration.co_names):
                try:
                    return _names.index(attribute_name)
                except ValueError:
                    if attribute_name.startswith('_'):
                        # we don't care about the order of magic and non
                        # public attributes
                        return 0
                    else:
                        msg = ("Can't find {0} in {1} class declaration. "
                                .format(attribute_name,
                                        class_declaration.co_name))
                        msg += ("HINT: use spyne.util.odict.odict for "
                                "class attributes if you populate them"
                                " dynamically.")
                        raise AttributeNotFoundException(msg)

            by_appearance = sorted(
                attributes.items(), key=lambda item: get_index(item[0])
            )

            namespace = cls.__prepare__(name, bases)
            for key, value in by_appearance:
                namespace[key] = value

            new_cls = constructor(cls, name, bases, namespace)

            found_module = inspect.getmodule(class_declaration)
            assert found_module is not None, (
                'Module is not found for class_declaration {0}, name {1}'
                .format(class_declaration, name))
            assert found_module.__name__ == new_cls.__module__, (
                'Found wrong class declaration of {0}: {1} != {2}.'
                .format(name, found_module.__name__, new_cls.__module__))

            return new_cls

        try:
            attributes["__new__"] = wraps(constructor)(preparing_constructor)
        except:
            warn("Wrapping class initializer failed. This is normal "
                          "when running under Nuitka")

        return type.__new__(cls, name, bases, attributes)
